/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

// TruePoly.  Find polymorphisms and non-polymorphisms in finished strain REFA vs 
// finished strain REFB.  Takes fasta or fastb files as input.
//
// Fix K.  A polymorphism between REFA and REFB is defined by sequences XAY
// in REFA and XBY in REFB where:
// - X and Y both have length K, and each occur exactly once in both REFA and REFB;
// - A and B are not both empty;
// - K-mers in XA other than X occur exactly once in REFA, and don't occur in REFB;
// - K-mers in XB other than X occur exactly once in REFB, and don't occur in REFA;
// - K-mers in AY other than Y occur exactly once in REFA, and don't occur in REFB;
// - K-mers in BY other than Y occur exactly once in REFB, and don't occur in REFA.
//
// There are several very slightly different but similar conditions that could be
// be used in place of the onces we use here.  It probably doesn't make much
// difference.
//
// If the polymorphism event is not a SNP or an indel, we split it up into several 
// such events.  After doing this, the flanking sequences match REFB, but do not 
// always match REFA.

// The output format is intended as a general format for output of polymorphism
// detection algorithms that compare sequence to a reference:
//
// line 1 = #name_of_reference_file (means REFB for this code)
// lines 2...n = fasta headers for reference file.
//
// Then one line per event with the following white-space-separated fields:
// [1] index of reference record (zero-based)
// [2] zero-based start of field [5] on reference record
// [3] left=... (left-flanking sequence on reference, K bases)
// [4] sample=... (sample sequence for polymorphism event, A as above)
// [5] reference=... (reference sequence for polymorphism event, B as above)
// [6] right=... (right-flanking sequence on reference, K bases).
// In the case of a SNP, both reference and sample will be one base long.  For an
// indel, one of reference or sample will be empty.  
//
// Last lines = summary:
// #n1 SNPs
// #n2 deletions from REFA
// #n3 deletions from REFB
//
// If TRUSTEDA is provided, it should be the name of a vecbitvector file matching
// REFA, and indicating which bases are trusted.
//
// A non-polymorphism between REFA and REFB is defined by sequences XAY
// in REFA and XBY in REFB where:
// - X and Y both have length K;
// - A and B both have length one, and are equal;
// - every K-mer in XAY = XBY occurs exactly once in both REFA and REFB.
// Under such circumstances, we label base B as a non-polymorphism base on REFB.
// In addition, bases lying in the flanking sequences X and Y for a polymorphism
// are also labeled as non-polymorphism bases on REFB.
//
// Note that in the event of an insertion in the sample, the reference bases
// at the locus may all be labeled non-polymorphism bases.  Note also that in a
// composite polymorphism, the internal bases that agree with the reference will
// not be labeled non-polymorphism bases.

#include "MainTools.h"
#include "polymorphism/TruePolyCore.h"

int main( int argc, char *argv[] ) 
{
     RunTime( );

     BeginCommandArguments;
     CommandArgument_String(REFA);
     CommandArgument_String(REFB);
     CommandArgument_String_OrDefault(TRUSTEDA, "");
     CommandArgument_String_OrDefault(TRUSTEDB, "");
     CommandArgument_Int_OrDefault(K, 28);
     CommandArgument_String_OrDefault(CLASSES_TO_PRINT, "{1,2,3}");
     CommandArgument_String_OrDefault(BINARY_FILE, "");
     CommandArgument_String_OrDefault(COMPOUND_FILE, "");
     EndCommandArguments;

     TruePolyCore( cout, REFA, REFB, TRUSTEDA, TRUSTEDB, K, CLASSES_TO_PRINT, 
          BINARY_FILE, COMPOUND_FILE );    }
