/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

/// VAAL.  Run VAAL{0,0.5,1,2,3,4} in succession.
/// See VAAL1.cc for documentation.

// MakeDepend: dependency VAAL0
// MakeDepend: dependency VAAL0_5
// MakeDepend: dependency VAAL1
// MakeDepend: dependency VAAL1_5
// MakeDepend: dependency VAAL2
// MakeDepend: dependency VAAL3
// MakeDepend: dependency VAAL4
// MakeDepend: dependency MakeLookupTable
// MakeDepend: dependency QueryLookupTable
// MakeDepend: dependency ShortQueryLookup
// MakeDepend: dependency ImperfectLookupTable

#include <strstream>

#include "Basevector.h"
#include "MainTools.h"
#include "math/Functions.h"
#include "math/HoInterval.h"
#include "micro/VectorTrimmer.h"
#include "paths/KmerPath.h"
#include "paths/ReadsToPathsCoreX.h"
#include "polymorphism/VAALTools.h"
#include "solexa/SolexaTools.h"

int main(int argc, char *argv[]) {
     RunTime();

     BeginCommandArguments;
     CommandArgument_String_Doc(HEAD,
          "Common prefix for a given flowcell lane's files (fastb, qltout).");
     CommandArgument_String_Doc(REF,
          "The reference to which the the reads should be compared.");
     CommandArgument_String_OrDefault_Doc(REF_TRUSTED, "",
          "A vecbitvector indicating which of the bases in the reference can be trusted.");
     CommandArgument_String_OrDefault_Doc(TRUE_REF, "",
          "A 'true' reference, if available, to check how well we did.");
     CommandArgument_Bool_OrDefault(SELF, False);
     CommandArgument_Int_OrDefault_Doc(TRUNC, 0, "Truncate the reads beyond this "
          "length (or don't truncate at all if TRUNC==0).");
     CommandArgument_String_OrDefault(RUN, "");
     CommandArgument_String_OrDefault(DIR, "");
     CommandArgument_String_OrDefault(POST, "");
     CommandArgument_Int_OrDefault(ADD_PASSES, -1);
     CommandArgument_Double_OrDefault_Doc(START, 0, "point in the sequence"
               " of programs to run where we should begin");
     CommandArgument_Double_OrDefault_Doc(STOP, 4, "point in the sequence of "
               "programs we should end");
     CommandArgument_Int_OrDefault(QMIN, -1);
     CommandArgument_String_OrDefault(PIPELINE, "");
     CommandArgument_Bool_OrDefault(COMPOUND, True);
     CommandArgument_Bool_OrDefault_Doc(UNIQUE_ALIGN, False, "Align faster, but lose polymorphisms.");
     CommandArgument_Bool_OrDefault_Doc(ILT_ALIGN, False, "Align even faster, but lose more polymorphisms.  Overrides UNIQUE_ALIGN.");
     CommandArgument_String_OrDefault_Doc(VECSEQ,VectorTrimmer::getDefaultSolexaSequence(),"vector sequence to be trimmed from right side of reads");
     EndCommandArguments;
     if ( PIPELINE == "" ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

     if (UNIQUE_ALIGN || ILT_ALIGN)
     {    cout << "Warning.  Using UNIQUE_ALIGN or ILT_ALIGN = True.  \n"
               << "This will lose some polymorphism calls, though the run \n"
               << "will be faster.  Particularly lose many short indels.\n"
               << "Cancel this run if this was not intentional.\n";    }

     if (SELF) { TRUE_REF = REF; }

     String workdir, workhead;
     String REFY = REF, DIRY = DIR, POSTY = POST;
     VAALSetup( RUN, HEAD, REFY, DIRY, POSTY, workdir, workhead,
          TRUNC, PIPELINE );
     if ( START == 0 ) {
          cout << workhead << endl;
          Ofstream( out, workhead + ".VAAL.out" );
          command.PrintTheCommandPretty( out );
     }

     // Construct the command-lines for VAAL{0-4}.

     for ( double N = START; N <= STOP; N += 0.5 ) {
          if ( N == 2.5 || N == 3.5 ) { continue; }

          String out = ToString(N);
          String::size_type pos = 0;
          while ((pos = out.find(".", pos)) != String::npos) 
          { out.replace(pos, 1, "_"); pos++; }

          String command = "VAAL" + out;
          command += " HEAD=\"" + HEAD + "\"";
          command += " PIPELINE=" + PIPELINE + " REF=" + REF;

          if ( N == 1 || N == 1.5 || N == 2 || N == 3 ) command += ".edit1";
          if ( TRUNC != 0 ) command += " TRUNC=" + ToString(TRUNC);
          if ( DIR != "" ) command += " DIR=" + DIR;
          if ( POST != "" ) command += " POST=" + POST;
          if ( RUN != "" && N != 0 ) command += " RUN=" + RUN;
          if ( N == 3 && QMIN >= 0 ) command += " QMIN=" + ToString(QMIN);
          if ( N == 0.5 && QMIN >= 0 ) command += " QMIN=" + ToString(QMIN);
          if ( N == 0 ) command += " VECSEQ=" + VECSEQ;
          if ( N == 1 && ADD_PASSES != -1 )
               command += " ADD_PASSES=" + ToString(ADD_PASSES);
          if ( (N == 0 || N == 0.5) && UNIQUE_ALIGN )
               command += " UNIQUE_ALIGN=True";
          if ( (N == 0 || N == 0.5) && ILT_ALIGN )
               command += " ILT_ALIGN=True";
          if ( N <= 3 ) SystemSucceed(command);
          if ( N == 4 ) {
               if ( COMPOUND ) { command += " COMPOUND=True"; }
               if ( REF_TRUSTED != "" ) { command += " REF_TRUSTED=" + REF_TRUSTED; }
               if ( TRUE_REF != "" ) { command += " TRUE_REF=" + TRUE_REF; }
               /*cout << N << " " << (command + " K=20") << endl;
               cout << N << " " << (command + " K=28") << endl;
               cout << N << " " << (command + " K=40") << endl;*/
               SystemSucceed( command + " K=20" );
               SystemSucceed( command + " K=28" );
               SystemSucceed( command + " K=40" );
          }
     }
}
