/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

// VAAL0.  Generate vector-trimmed reads and align them.  Called by VAAL.
// See VAAL1.cc for general documentation.


#include "MainTools.h"
#include "micro/VectorTrimmer.h"
#include "micro/ArtifactualSequenceDetector.h"
#include "solexa/SolexaTools.h"
#include "feudal/HashSet.h"
#include "feudal/VirtualMasterVec.h"
#include "Basevector.h"
#include "Qualvector.h"
#include "solexa/FourBase.h"
#include "feudal/IncrementalWriter.h"
#include <cstring>
#include <functional>

namespace
{
    bool needsUpdate( String const& oldFile, String const& newFile, bool isFeudal = true )
    {
        bool result = false;
        if ( !IsRegularFile(newFile) || (isFeudal && !IsGoodFeudalFile(newFile)) )
        {
            if ( !IsRegularFile(oldFile) )
                FatalErr("Can't produce " + newFile + " because " + oldFile + " does not exist.");
            result = true;
        }
        else if ( IsRegularFile(oldFile) && IsOlder(newFile,oldFile) )
        {
            result = true;
        }
        return result;
    }

    class OddMer
    {
    public:
        OddMer() : mVal(0) {}

        // compiler-supplied copying and destructor are OK

        OddMer next( unsigned char base ) const
        { OddMer result(*this);
          result.mVal = (result.mVal << 2) | base;
          return result; }

        // unusual, but fast, canonicalization: minimize the middle base
        // (since K is odd, there will always be a middle base)
        OddMer canonicalize() const
        { return (mVal & (3 << (K&~1))) <= (1 << (K&~1)) ? *this : rc(); }

        OddMer rc() const
        { OddMer result(*this); Base::rc(result.mVal);
          result.mVal >>= ((32-K)*2); return result; }

        unsigned long hash() const
        { unsigned long result = 14695981039346656037ul;
          unsigned char const* itr(reinterpret_cast<unsigned char const*>(&mVal));
          unsigned char const* end(itr+sizeof(mVal));
          for ( ; itr != end; ++itr )
              result = 1099511628211ul*(result ^ *itr); // FNV-1a algorithm
          return result; }

        struct Hasher : public std::unary_function<OddMer,unsigned long>
        {
          unsigned long operator()( OddMer const& om ) const
          { return om.hash(); }
        };

        friend bool operator==( OddMer const& v1, OddMer const& v2 )
        { return v1.mVal == v2.mVal; }

        static unsigned int const K = 21; // must be odd and < 32

    private:
        unsigned long mVal;
    };

    class FreqCount
    {
    public:
        FreqCount() : mGenomeCount(0), mReadCount(0) {}

        // compiler-supplied copying and destructor are OK

        unsigned int genomeCount() const { return mGenomeCount; }
        unsigned int readCount() const { return mReadCount; }
#ifdef NDEBUG
        void incrementGenomeCount() { mGenomeCount += 1; }
        void incrementReadCount() { mReadCount += 1; }
#else
        void incrementGenomeCount()
        { if ( (mGenomeCount += 1) > mMaxGenomeCount )
            mMaxGenomeCount = mGenomeCount; }

        void incrementReadCount()
        { if ( (mReadCount += 1) > mMaxReadCount )
            mMaxReadCount = mReadCount; }

        static unsigned int maxGenomeCount() { return mMaxGenomeCount; }
        static unsigned int maxReadCount() { return mMaxReadCount; }

    private:
        static unsigned int mMaxGenomeCount;
        static unsigned int mMaxReadCount;
#endif
    private:
        unsigned int mGenomeCount;
        unsigned int mReadCount;
    };

#ifndef NDEBUG
    unsigned int FreqCount::mMaxGenomeCount = 0;
    unsigned int FreqCount::mMaxReadCount = 0;
#endif

    class KmerCounter
    {
    public:
        KmerCounter() : mMap(INITIAL_MAP_SIZE) {}

        // compiler-supplied copying and destructor are OK

        void countGenomicKmers( String const& refFastaFile );
        bool isOverrepresented( bvec const& bv );
        std::size_t getNKMers() { return mMap.size(); }

    private:

        struct Entry : public OddMer
        {
            Entry() {}
            Entry( OddMer const& oddmer ) : OddMer(oddmer) {}

            // compiler-supplied copying and destructor is OK

            FreqCount mFreq;
        };
        HashSet<Entry,OddMer::Hasher,std::equal_to<OddMer> > mMap;

        static std::size_t const INITIAL_MAP_SIZE = 100000000;
        static unsigned int const MAX_REPS = 100;
    };

    void KmerCounter::countGenomicKmers( String const& refFastaFile )
    {
        std::ifstream is(refFastaFile.c_str());
        char buf[8192];
        OddMer kmer;
        std::size_t nBasesInKmer = 0;
        while ( is.getline(buf,sizeof(buf)) )
        {
            if ( buf[0] == '>' )
            {
                kmer = OddMer();
                nBasesInKmer = 0;
            }
            else
            {
                char* end = buf + strlen(buf);
                for ( char* itr(buf); itr != end; ++itr )
                {
                    if ( !Base::isBase(*itr) )
                    {
                        kmer = OddMer();
                        nBasesInKmer = 0;
                    }
                    else
                    {
                        kmer = kmer.next(Base::char2Val(*itr));
                        if ( ++nBasesInKmer >= OddMer::K )
                        {
                            Entry const& entry = mMap[kmer.canonicalize()];
                            FreqCount& fc = const_cast<FreqCount&>(entry.mFreq);
                            fc.incrementGenomeCount();
                        }
                    }
                }
            }
        }
        if ( !is.eof() )
            FatalErr("Failed in reading reference file " << refFastaFile);
    }

    bool KmerCounter::isOverrepresented( bvec const& bv )
    {
        typedef bvec::const_iterator BVCI;

        bool result = false;
        if ( bv.size() >= OddMer::K )
        {
            OddMer oddMer;

            BVCI itr(bv.begin());
            for ( BVCI end(bv.begin(OddMer::K-1)); itr != end; ++itr )
                oddMer = oddMer.next(*itr);

            for ( BVCI end(bv.end()); itr != end; ++itr )
            {
                oddMer = oddMer.next(*itr);
                Entry const& entry = mMap[oddMer.canonicalize()];
                FreqCount& fc = const_cast<FreqCount&>(entry.mFreq);
                fc.incrementReadCount();

                // make a fake genome count of 1 for kmers unique to the reads
                if ( !fc.genomeCount() ) fc.incrementGenomeCount();

                if ( fc.readCount()/fc.genomeCount() >= MAX_REPS )
                    result = true;
            }
        }
        return result;
    }
}

int main( int argc, char *argv[] )
{
    RunTime();
    BeginCommandArguments;
    CommandArgument_String_Doc(HEAD,
            "The common file prefix for a flowcell lane's files");
    CommandArgument_String_Doc(REF, "The reference");
    CommandArgument_Int_OrDefault_Doc(TRUNC, 0,
            "Truncate the reads to this length");
    CommandArgument_String_OrDefault_Doc(PIPELINE, "",
            "The location of the pipelineOutput directory");
    CommandArgument_Bool_OrDefault_Doc(UNIQUE_ALIGN, False,
            "Go faster, but lose polymorphisms.");
    CommandArgument_Bool_OrDefault_Doc(ILT_ALIGN, False,
            "Go even faster, but lose more polymorphisms.");
    CommandArgument_String_OrDefault_Doc(VECSEQ,VectorTrimmer::getDefaultSolexaSequence(),
            "Vector sequence to be trimmed from right end of reads.");
    CommandArgument_String_OrDefault_Doc(ARTIFACTS,"",
            "Fastb file of artifactual sequences to use in rejecting reads.  Defaults to standard Solexa sequences.");
    EndCommandArguments;
    if ( PIPELINE == "" ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

    vec<String> HEADS;
    ExpandHead(HEAD, HEADS, TRUNC, PIPELINE); // Perform the brace expansion that the shell doesn't do for "KEY=VALUE"-like arguments.

    // Infer the location of the reference file
    REF += ".fasta";
    String REFX = REF;
    String run = HEADS[0].RevBefore("/");
    if ( IsRegularFile(run + "/" + REF) )
        REFX = run + "/" + REF;
    else if ( run.Contains("/") )
    {
        run = run.RevBefore("/");
        if ( IsRegularFile(run + "/" + REF) )
            REFX = run + "/" + REF;
    }

    // Infer the location of the lookup file
    String lookup = REFX.Before(".fasta") + ".lookup";
    if ( needsUpdate(REFX, lookup, false) )
        SystemSucceed("MakeLookupTable SOURCE=" + REFX + " OUT_HEAD=" + REFX.Before(".fasta") + " LOOKUP_ONLY=True");

    String qltoutSuffix = ".vectrim.qltout." + REF.Before(".fasta");
    VectorTrimmer trimmer(VECSEQ);
    ArtifactualSequenceDetector asd(ARTIFACTS.c_str());
    KmerCounter kmerCounter;

    kmerCounter.countGenomicKmers( REFX );

    // Loop over all the lanes specified by the user
    size_t nReads = 0;
    size_t nArtifactualReads = 0;
    size_t nOverrepresentedReads = 0;
    for ( int i = 0; i < HEADS.isize(); i++ )
    {
        String fastb = HEADS[i] + ".fastb";
        String vectrimFastb = HEADS[i] + ".vectrim.fastb";
        String intensities = HEADS[i] + ".intensities";
        String vectrimIntensities = HEADS[i] + ".vectrim.intensities";
        String qualb = HEADS[i] + ".qualb";
        String vectrimQualb = HEADS[i] + ".vectrim.qualb";
        bool intensitiesExist = IsRegularFile(intensities);

        if ( needsUpdate(fastb, vectrimFastb) || needsUpdate(qualb,vectrimQualb)
                || (intensitiesExist && needsUpdate(intensities,vectrimIntensities)) )
        {
            VirtualMasterVec<bvec> vbv(fastb.c_str());
            VirtualMasterVec<qvec> vqv(qualb.c_str());
            if ( vbv.size() != vqv.size() )
                FatalErr("Sequence file " + fastb + " and quality file " + qualb + " have different sizes.");

            IncrementalWriter<bvec> bw(vectrimFastb.c_str(), vbv.size());
            IncrementalWriter<qvec> qw(vectrimQualb.c_str(), vbv.size());
            VirtualMasterVec<bvec>::Itr end(vbv.end());
            VirtualMasterVec<bvec>::Itr bItr(vbv.begin());
            VirtualMasterVec<qvec>::Itr qItr(vqv.begin());

            if ( !intensitiesExist )
            {
                while ( bItr != end )
                {
                    nReads += 1;
                    bvec bv(*bItr);
                    trimmer.trim(bv);
                    if ( asd.isArtifactual(bv) )
                    {
                        bv.resize(0);
                        nArtifactualReads += 1;
                    }
                    else if ( kmerCounter.isOverrepresented(bv) )
                    {
                        bv.resize(0);
                        nOverrepresentedReads += 1;
                    }
                    bw.add(bv);

                    qvec qv(*qItr);
                    qv.resize(bv.size());
                    qw.add(qv);

                    ++bItr;
                    ++qItr;
                }
            }
            else
            {
                VirtualMasterVec<FourBaseVec> viv(intensities.c_str());
                if ( vbv.size() != viv.size() )
                    FatalErr("Sequence file " + fastb +
                             " and intensities file " + intensities +
                             " have different sizes.");

                IncrementalWriter<FourBaseVec> iw(vectrimIntensities.c_str(),
                                                    vbv.size());
                VirtualMasterVec<FourBaseVec>::Itr iItr(viv.begin());
                while ( bItr != end )
                {
                    nReads += 1;
                    bvec bv(*bItr);
                    trimmer.trim(bv);
                    if ( asd.isArtifactual(bv) )
                    {
                        bv.resize(0);
                        nArtifactualReads += 1;
                    }
                    else if ( kmerCounter.isOverrepresented(bv) )
                    {
                        bv.resize(0);
                        nOverrepresentedReads += 1;
                    }
                    bw.add(bv);

                    qvec qv(*qItr);
                    qv.resize(bv.size());
                    qw.add(qv);

                    FourBaseVec iv(*iItr);
                    iv.resize(bv.size());
                    iw.add(iv);

                    ++bItr;
                    ++qItr;
                    ++iItr;
                }
                iw.close();
            }
            bw.close();
            qw.close();
        }

        // Search the qltout file.  If it's complete (as indicated by the presence of the phrase ": done."), then move on.  Otherwise, regenerate it.
        String outfile = HEADS[i] + qltoutSuffix;
        if ( needsUpdate(vectrimFastb, outfile, false) ||
                (!LineOfOutput("tail -1 " + outfile).Contains(": done.", -1)
                 && !LineOfOutput("tail -2 " + outfile + " | head -1").Contains(": done.", -1)) )
        {
            String args = " ERR_DIFF=2 MAX_ERRS=4 CHUNK=500000 SEQS=" + vectrimFastb +
                            " L=" + lookup + " O=" + HEADS[i] + " OUT_SUFFIX=" + qltoutSuffix;
            // Guess our qltout file was either missing, stale, or incomplete.  Let's regenerate it!
            if ( ILT_ALIGN )
                SystemSucceed("ImperfectLookupTable" + args);
            else
            {
                if ( UNIQUE_ALIGN )
                    args += " UNIQUE_ONLY=True";
                SystemSucceed("ShortQueryLookup" + args);
            }
        }
    }

    std::cout << "VAAL0 Processed " << nReads << " reads." << std::endl;
    std::cout << "VAAL0 Truncated " << nArtifactualReads << " reads as artifactual." << std::endl;
    std::cout << "VAAL0 Truncated " << nOverrepresentedReads << " reads as redundant." << std::endl;
    std::cout << "There are " << kmerCounter.getNKMers() << " kmers." << std::endl;
#ifndef NDEBUG
    std::cout << "The maximum genomic frequency is " << FreqCount::maxGenomeCount() << std::endl;
    std::cout << "The maximum read frequency is " << FreqCount::maxReadCount() << std::endl;
#endif
}
