/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

// VAAL0.5.  Use sample reads to edit reference, using approximately the
// following scheme:
// 1. Align reads to reference.  (Currently, these are provided in advance.)
// 2. Pick unique alignments.
// 3. Using base quality, alignments define list of somewhat trusted events on
//    reference (mutations or very small indels).
// 4. Define window around each event.  (Proximate events go into same window.)
// 5. Pick the reads that aligned uniquely to that window.
// 6. If there are n events, consider all 2^n possible states of the window:
//    each event can be either on or off.  (Note n should be very small.)  Form
//    mutated reference corresponding to each state.  Realign the local reads to the
//    window and create composite score.  Pick the state yielding the lowest
//    score and edit the reference.
// 7. Exit if tired.  (Currently, we always exit.)
// 8. Goto 1.

#include "Alignment.h"
#include "Basevector.h"
#include "MainTools.h"
#include "lookup/LookAlign.h"
#include "math/Functions.h"
#include "pairwise_aligners/SmithWatFree.h"
#include "solexa/FourBase.h"
#include "solexa/SolexaMetrics.h"
#include "polymorphism/VAALTools.h"
#include "solexa/SolexaTools.h"

class poly_event {

     public:

     enum ConstructorBehavior { MUTATION, DELETION, INSERTION };

     poly_event( ) { }

     poly_event( const int tig, const int start, const char base,
          const ConstructorBehavior constructor_type )
          : tig_(tig), start_(start)
     {    ForceAssertEq( (int)constructor_type, (int)MUTATION );
          poly_type_ = 0;
          bases_.resize(1);
          bases_.Set( 0, base );    }

     poly_event( const int tig, const int start, const int del,
          const ConstructorBehavior constructor_type )
          : tig_(tig), start_(start)
     {    ForceAssertEq( (int)constructor_type, (int)DELETION );
          ForceAssertGt( del, 0 );
          poly_type_ = 1;
          del_size_ = del;    }

     poly_event( const int tig, const int start, const basevector insertion,
          const ConstructorBehavior constructor_type )
          : tig_(tig), start_(start)
     {    ForceAssertEq( (int)constructor_type, (int)INSERTION );
          ForceAssertGt( insertion.size( ), 0u );
          poly_type_ = 2;
          bases_ = insertion;    }

     int Tig( ) const { return tig_; }
     int Start( ) const { return start_; }
     int Stop( ) const
     {    if ( IsDeletion( ) ) return Start( ) + DeletionSize( );
          else return Start( ) + 1;    }

     char MutatedTo( ) const
     {    ForceAssert( poly_type_ == 0 );
          return bases_[0];    }

     int DeletionSize( ) const
     {    ForceAssert( poly_type_ == 1 );
          return del_size_;    }

     const basevector& Insertion( ) const
     {    ForceAssert( poly_type_ == 2 );
          return bases_;    }

     Bool IsMutation( ) const { return poly_type_ == 0; }
     Bool IsDeletion( ) const { return poly_type_ == 1; }
     Bool IsInsertion( ) const { return poly_type_ == 2; }

     friend Bool operator<( const poly_event& p1, const poly_event& p2 )
     {    if ( p1.Tig( ) < p2.Tig( ) ) return True;
          if ( p1.Tig( ) > p2.Tig( ) ) return False;
          if ( p1.Start( ) < p2.Start( ) ) return True;
          if ( p1.Start( ) > p2.Start( ) ) return False;
          if ( p1.poly_type_ < p2.poly_type_ ) return True;
          if ( p1.poly_type_ > p2.poly_type_ ) return False;
          if ( p1.del_size_ < p2.del_size_ ) return True;
          if ( p1.del_size_ > p2.del_size_ ) return False;
          if ( p1.bases_ < p2.bases_ ) return True;
          return False;    }

     friend Bool operator==( const poly_event& p1, const poly_event& p2 )
     {    return p1.Tig( ) == p2.Tig( ) && p1.Start( ) == p2.Start( )
               && p1.poly_type_ == p2.poly_type_ && p1.del_size_ == p2.del_size_
               && p1.bases_ == p2.bases_;    }

     friend Bool operator!=( const poly_event& p1, const poly_event& p2 )
     {    return !( p1 == p2 );    }

     friend ostream& operator<<( ostream& out, const poly_event& p )
     {    out << p.Tig( ) << "." << p.Start( ) << " ";
          if ( p.IsMutation( ) ) out << as_base( p.MutatedTo( ) );
          if ( p.IsDeletion( ) ) out << "delete " << p.del_size_ << " bases";
          if ( p.IsInsertion( ) ) out << "insert " << p.bases_.ToString( );
          return out << "\n";    }

     private:

     int tig_;
     int start_;
     int poly_type_; // 0 = mutation, 1 = deletion, 2 = insertion
     int del_size_;
     basevector bases_;

};

class edit_window {

     public:

     edit_window( ) { }

     edit_window( const int tig, const int start, const int stop,
          const vec<poly_event>& polys )
          : tig_(tig), start_(start), stop_(stop), polys_(polys)
     {    ForceAssertLe( start, stop );    }

     int Tig( ) const { return tig_; }
     int Start( ) const { return start_; }
     int Stop( ) const { return stop_; }
     int Length( ) const { return Stop( ) - Start( ); }
     const vec<poly_event>& Polys( ) const { return polys_; }
     int NPolys( ) const { return polys_.size( ); }
     const poly_event& Poly( int i ) const { return polys_[i]; }
     const vec<int>& ReadIds( ) const { return read_ids_; }
     int ReadId( int i ) const { return read_ids_[i]; }
     int NReads( ) const { return read_ids_.size( ); }
     Bool Fw( int i ) const { return !read_dirs_[i]; }
     Bool Rc( int i ) const { return read_dirs_[i]; }

     void AddRead( const int id, const Bool rc )
     {    read_ids_.push_back(id);
          read_dirs_.push_back(rc);    }

     friend ostream& operator<<( ostream& out, const edit_window& w )
     {    out << "\nwindow " << w.Tig( ) << "." << w.Start( ) << "-" << w.Stop( )
               << ", " << w.NReads( ) << " reads\n";
          for ( int i = 0; i < w.NPolys( ); i++ )
               out << w.Poly(i);
          return out;    }

     private:

     int tig_;
     int start_, stop_;
     vec<poly_event> polys_;
     vec<int> read_ids_;
     vec<Bool> read_dirs_;

};

int main( int argc, char *argv[] )
{
     RunTime( );

     BeginCommandArguments;
     CommandArgument_String(HEAD);
     CommandArgument_String(REF);
     CommandArgument_String_OrDefault( RUN, "" );
     CommandArgument_Int_OrDefault(TRUNC, 0);
     CommandArgument_String_OrDefault(DIR, "");
     CommandArgument_String_OrDefault(POST, "");
     CommandArgument_String_OrDefault(PIPELINE, "");
     CommandArgument_Bool_OrDefault(UNIQUE_ALIGN, False);
     CommandArgument_Bool_OrDefault(ILT_ALIGN, False);
     CommandArgument_Int_OrDefault_Doc(QMIN, 10,
          "Define trusted read bases to be those having quality score >= QMIN." );
     EndCommandArguments;
     if ( PIPELINE == "" ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

     // Define directories, etc.

     String workdir, workhead;
     VAALSetup( RUN, HEAD, REF, DIR, POST, workdir, workhead, TRUNC,
          PIPELINE );
     cout << "writing to " << workhead + ".SolexaEdit.out\n" << endl;
     Ofstream( out, workhead + ".SolexaEdit.out" );
     command.PrintTheCommandPretty( out );

     // Load data.

     vecbasevector bases, ref;
     String REFX;
     VAALLoad( workhead, HEAD, ".vectrim", TRUNC, REF, REFX, bases,
          ref, PIPELINE );
     int nreads = bases.size( );
     vecqualvector Q;
     vec<String> HEADS;
     ExpandHead( HEAD, HEADS, TRUNC, PIPELINE );
     for ( int i = 0; i < HEADS.isize( ); i++ )
          Q.ReadAll( HEADS[i] + ".qualb", i > 0 );

     // Find putative polymorphism events.

     int readcount = 0, nplaced1p = 0;
     vec<poly_event> polys;
     for ( int i = 0; i < HEADS.isize( ); i++ )
     {    String HEAD = HEADS[i] + ".vectrim";
          vec<look_align> aligns;
          vec< vec<int> > aligns_index;
          vecbasevector bases( HEAD + ".fastb" );
          vecbasevector basesrc(bases);
          ReverseComplement(basesrc);
          int nreadsi = bases.size( );
          LoadLookAligns( HEAD + ".qltout." + REF.Before( ".fasta" ),
               aligns, aligns_index, nreadsi );
          for ( int jid = 0; jid < nreadsi; jid++ )
          {    int id = jid + readcount;
               int min_errors = 1000000000;
               for ( int u = 0; u < aligns_index[jid].isize( ); u++ )
               {    const look_align& la = aligns[ aligns_index[jid][u] ];
                    min_errors = Min( min_errors, la.Errors( ) );    }
               if ( min_errors == 1000000000 ) continue;
               static vec<int> at_min;
               at_min.clear( );
               for ( int u = 0; u < aligns_index[jid].isize( ); u++ )
               {    const look_align& la = aligns[ aligns_index[jid][u] ];
                    if ( la.Errors( ) == min_errors ) at_min.push_back(u);    }
               if ( !at_min.solo( ) ) continue;
               const look_align& la = aligns[ aligns_index[jid][ at_min[0] ] ];
               int tig = la.target_id, p1 = la.pos1( ), p2 = la.pos2( );
               if ( la.Pos2( ) > ref[tig].isize( ) )
               {    cout << "Alignment of read " << jid << " from "
                         << HEAD << " appears to be inconsistent with the "
                         << "reference that was used.\n";
                    cout << "Abort.\n";
                    exit(1);    }
               int n = bases[jid].size( );
               static qualvector q;
               q = Q[jid];
               if (la.rc1) q.ReverseMe( );
               for ( int j = 0; j < la.a.Nblocks( ); j++ )
               {    if ( la.a.Gaps(j) > 0 )
                    {    Bool good = True;
                         for ( int x = 0; x <= 1; x++ )
                         {    int p1x = ( !la.rc1 ? p1+x : n - (p1+x) - 1 );
                              if ( p1x < 0 || p1x >= n ) continue;
                              if ( q[p1+x] < QMIN ) good = False;    }
                         if (good)
                         {    polys.push( tig, p2, la.a.Gaps(j),
                                   poly_event::DELETION );    }
                         p2 += la.a.Gaps(j);    }
                    if ( la.a.Gaps(j) < 0 )
                    {    Bool good = True;
                         for ( int x = 0; x < -la.a.Gaps(j); x++ )
                         {    int p1x = ( !la.rc1 ? p1+x : n - (p1+x) - 1 );
                              if ( q[p1+x] < QMIN ) good = False;    }
                         if (good)
                         {    basevector insertion( -la.a.Gaps(j) );
                              for ( int x = 0; x < -la.a.Gaps(j); x++ )
                              {    char base;
                                   if ( !la.rc1 ) base = bases[jid][p1+x];
                                   else base = basesrc[jid][p1+x];
                                   insertion.Set( x, base );    }
                              polys.push( tig, p2, insertion,
                                   poly_event::INSERTION );    }
                         p1 -= la.a.Gaps(j);    }
                    for ( int x = 0; x < la.a.Lengths(j); x++ )
                    {    int p1x = ( !la.rc1 ? p1 : n - p1 - 1 );
                         if ( q[p1] >= QMIN )
                         {    char base;
                              if ( !la.rc1 ) base = bases[jid][p1];
                              else base = basesrc[jid][p1];
                              if ( base != ref[tig][p2] )
                              {    polys.push( tig, p2, base,
                                        poly_event::MUTATION );    }    }
                         ++p1;
                         ++p2;    }    }    }
          readcount += bases.size( );    }

     // Discard events that occur less than three times.

     Sort(polys);
     vec<poly_event> polys2;
     for ( int i = 0; i < polys.isize( ); i++ )
     {    int j;
          for ( j = i + 1; j < polys.isize( ); j++ )
               if ( polys[j] != polys[i] ) break;
          if ( j - i >= 3 ) polys2.push_back( polys[i] );
          i = j - 1;    }
     polys = polys2;


     // We moved the ref2 here since we may completely skip
     // Assign edit windows etc. until the write output section
     vecbasevector ref2(ref);


     // Define windows.

     vec<edit_window> windows;
     int maxread = 0;
     for ( size_t i = 0; i < bases.size( ); i++ )
          maxread = Max( maxread, bases[i].isize( ) );
     for ( int i = 0; i < polys.isize( ); i++ )
     {    int tig = polys[i].Tig( );
          int j;
          for ( j = i + 1; j < polys.isize( ); j++ )
          {    if ( polys[j].Tig( ) != tig
                    || polys[j].Start( ) - polys[i].Stop( ) >= maxread ) break;    }
          int start = Max( 0, polys[i].Start( ) - maxread );
          int stop = Min( ref[tig].isize( ), polys[j-1].Stop( ) + maxread );
          windows.push( tig, start, stop, SubOf( polys, i, j - i ) );
          i = j - 1;    }
     if ( windows.empty( ) ) {
          cout << "Warning: no edit windows.\n";
     } else {

     // Assign reads to windows.

     vec< pair<int,int> > starts;
     for ( int i = 0; i < windows.isize( ); i++ )
          starts.push( windows[i].Tig( ), windows[i].Start( ) );
     readcount = 0;
     for ( int i = 0; i < HEADS.isize( ); i++ )
     {    String HEAD = HEADS[i] + ".vectrim";
          vec<look_align> aligns;
          vec< vec<int> > aligns_index;
          vecbasevector bases( HEAD + ".fastb" );
          vecbasevector basesrc(bases);
          ReverseComplement(basesrc);
          int nreadsi = bases.size( );
          // solexa_metric_db db( HEADS[i] + ".metrics" );
          LoadLookAligns( HEAD + ".qltout." + REF.Before( ".fasta" ),
               aligns, aligns_index, nreadsi );
          for ( int jid = 0; jid < nreadsi; jid++ )
          {    int id = jid + readcount;
               int min_errors = 1000000000;
               for ( int u = 0; u < aligns_index[jid].isize( ); u++ )
               {    const look_align& la = aligns[ aligns_index[jid][u] ];
                    min_errors = Min( min_errors, la.Errors( ) );    }
               if ( min_errors == 1000000000 ) continue;
               static vec<int> at_min;
               at_min.clear( );
               for ( int u = 0; u < aligns_index[jid].isize( ); u++ )
               {    const look_align& la = aligns[ aligns_index[jid][u] ];
                    if ( la.Errors( ) == min_errors ) at_min.push_back(u);    }
               if ( !at_min.solo( ) ) continue;
               const look_align& la = aligns[ aligns_index[jid][ at_min[0] ] ];
               int tig = la.target_id, p1 = la.pos1( ), p2 = la.pos2( );
               int wid = upper_bound( starts.begin( ), starts.end( ),
                    make_pair( tig, la.pos2( ) ) ) - starts.begin( ) - 1;
               wid = Max( 0, wid );
               if ( windows[wid].Start( ) <= la.pos2( )
                    && la.Pos2( ) <= windows[wid].Stop( ) )
               {    windows[wid].AddRead( id, la.rc1 );    }    }
          readcount += bases.size( );    }

     // Consider each possible edit of each window.

     vec<int> ref_offset( ref.size( ), 0 );
     for ( int i = 0; i < windows.isize( ); i++ )
     {    const edit_window& w = windows[i];
          out << w << "\n";
          const int max_polys_in_window = 7;
          int np = w.NPolys( );
          if ( np > max_polys_in_window )
          {    out << "Too many polys in window.  Ignoring.\n";
               continue;    }
          int top = 1;
          for ( int i = 0; i < np; i++ )
               top *= 2;
          basevector ref_orig;
          int tig = w.Tig( );
          ref_orig.SetToSubOf( ref[tig], w.Start( ), w.Length( ) );
          vec<int> errs( top, 0 );
          vec<basevector> ref_edit(top);
          for ( int z = 0; z < top; z++ )
          {    vec<Bool> edit( np, False );
               int x = z;
               for ( int j = 0; j < np; j++ )
               {    if ( x % 2 ) edit[j] = True;
                    x = x >> 1;    }
               out << "trying ";
               for ( int j = 0; j < np; j++ )
                    out << ( edit[j] ? "+" : "-" );

               // Compute ref_edit.

               ref_edit[z] = ref_orig;
               int offset = 0;
               for ( int j = 0; j < np; j++ )
               {    if ( !edit[j] ) continue;
                    const poly_event& p = w.Poly(j);
                    int start = offset + p.Start( ) - w.Start( );
                    if ( p.IsMutation( ) ) ref_edit[z].Set( start, p.MutatedTo( ) );
                    if ( p.IsDeletion( ) )
                    {    int del = p.DeletionSize( );
                         basevector b1, b2;
                         b1.SetToSubOf( ref_edit[z], 0, start );
                         b2.SetToSubOf( ref_edit[z],
                              start + del, ref_edit[z].isize( ) - start - del );
                         ref_edit[z] = Cat( b1, b2 );
                         offset -= del;    }
                    if ( p.IsInsertion( ) )
                    {    basevector b1, b2;
                         b1.SetToSubOf( ref_edit[z], 0, start );
                         b2.SetToSubOf( ref_edit[z], start,
                              ref_edit[z].isize( ) - start );
                         ref_edit[z] = Cat( b1, p.Insertion( ), b2 );
                         offset += p.Insertion( ).size( );    }    }

               // Align the reads to edited reference.

               for ( int i = 0; i < w.NReads( ); i++ )
               {    basevector r = bases[ w.ReadId(i) ];
                    if ( w.Rc(i) ) r.ReverseComplement( );
                    int best_loc;
                    static alignment a;
                    errs[z] += SmithWatFree( r, ref_edit[z], best_loc, a );    }
               out << " --> " << errs[z] << " errors\n";    }
          if ( Min(errs) >= errs[0] ) continue;
          int z;
          for ( z = 1; z < errs.isize( ); z++ )
               if ( errs[z] == Min(errs) ) break;
          out << "CHANGE\n";
          basevector r1, r2;
          r1.SetToSubOf( ref2[tig], 0, ref_offset[tig] + w.Start( ) );
          r2.SetToSubOf( ref2[tig], ref_offset[tig] + w.Stop( ),
               ref2[tig].isize( ) - ref_offset[tig] - w.Stop( ) );
          ref2[tig] = Cat( r1, ref_edit[z], r2 );
          ref_offset[tig] += ref_edit[z].isize( ) - ref_orig.isize( );    }

     } // End of "else" part of "if (windows.empty()) ..."

     // Write edited reference.

     String refhead = workhead + "." + REF.Before( ".fasta" ) + ".edit1";
     ref2.WriteAll( refhead + ".fastb" );
     Ofstream( fout, refhead + ".fasta" );
     for ( size_t i = 0; i < ref2.size( ); i++ )
          ref2[i].Print( fout, i );
     SystemSucceed( "MakeLookupTable SOURCE=\"" + refhead + ".fastb\" "
          + "LOOKUP_ONLY=True SIMPLE_SOURCE=True OUT_HEAD=\"" + refhead + "\"" );

     // Create new alignments.
     String unique_option = "";
     if (UNIQUE_ALIGN) unique_option = "UNIQUE_ONLY=True";

     for ( int i = 0; i < HEADS.isize( ); i++ )
     {    String HEAD = HEADS[i] + ".vectrim";
          Mkdir777(workhead);
          if (!ILT_ALIGN) {
              SystemSucceed( "ShortQueryLookup " + unique_option + " SEQS=" + HEADS[i] + ".vectrim.fastb L=\""
                  + refhead + ".lookup\" O=\"" + workhead + "/" + HEAD.RevAfter( "/" )
                  + "\" OUT_SUFFIX=" + ".qltout." + REF.Before( ".fasta" ) + ".edit1"
                  +  " ERR_DIFF=2 MAX_ERRS=4 CHUNK=500000" );
          } else {
              SystemSucceed( "ImperfectLookupTable  SEQS=" + HEADS[i] + ".vectrim.fastb L=\""
                  + refhead + ".lookup\" O=\"" + workhead + "/" + HEAD.RevAfter( "/" )
                  + "\" OUT_SUFFIX=" + ".qltout." + REF.Before( ".fasta" ) + ".edit1"
                  +  " ERR_DIFF=2 MAX_ERRS=4 CHUNK=500000" );
          }
     }
}
