/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

// VAAL1.  This is part of a suite of programs VAAL{0,1,2,3,4} that are run to find
// polymorphisms of reads versus a given resistance, via an assisted assembly.
//
// This program generates files:
//
// * .marks file ("alignments of reads to reference")
// * .bases_ref.paths.20
// * .bases_ref.pathsrc.20
// * .trim_exempt.
//
// Shared requirements for VAAL{0,1,2,3,4}:
//
// 1. Run with HEAD=FC.LANE
//             or   DATE_FC.LANE
//             or   FC.LANE[TRUNC]
//             or   DATE_FC.LANE[TRUNC]
// where TRUNC is an integer; the directory trunc_TRUNC must exist and be
// appropriately populated.  Alternatively, TRUNC may be explicitly specified.
//
// 2. HEAD may also be a list like {head1,...,headn}.
//
// 3. REF.fasta and REF.lookup must exist in the run directory.  If there is more
// than one flowcell, we look in the first directory.
//
// 4. DIR and POST may be used to specify where the output and intermediary
// files should go.  Otherwise we make up something sensible.  If lanes do not come
// from one flowcell, DIR is set to /broad/solexaproc/pipelineOutput/mixed.
//
// 5. The variable RUN may be specified.  The default value is your user name.
// It is made part of the path.  This feature reduces the likelihood of
// accidental overwrites.

#include "Basevector.h"
#include "Bitvector.h"
#include "FeudalMimic.h"
#include "Intvector.h"
#include "MainTools.h"
#include "STLExtensions.h"
#include "feudal/BinaryStream.h"
#include "lookup/LookAlign.h"
#include "lookup/PerfectLookup.h"
#include "math/Functions.h"
#include "math/HoInterval.h"
#include "paths/KmerBaseBroker.h"
#include "paths/KmerPath.h"
#include "paths/Unipath.h"
#include "paths/ReadsToPathsCoreX.h"
#include "polymorphism/VAALTools.h"
#include "random/SequenceEntropy.h"
#include "solexa/SolexaTools.h"
#include <algorithm>

// Correct errors in the reads.
void correctErrors( String const& HEAD, int TRUNC, String const& PIPELINE,
                    String const& REF, String const& workhead,
                    vecbvec const& ref, vecbvec& bases_ref )
{
    vec<String> HEADS;
    ExpandHead(HEAD, HEADS, TRUNC, PIPELINE);
    vec<VecIntVec> calls(6);
    for ( int i = 0; i < 6; i++ )
        Mimic(ref, calls[i]);
    size_t totCoverage = 0;
    vecbasevector ref_fixed(ref);
    vecbitvector known;
    Mimic(ref, known);

    int readcount = 0;

    // pass 1
    for ( int i = 0; i < HEADS.isize(); i++ )
    {
        vec<look_align> aligns;
        vec<vec<int> > aligns_index;
        int nreadsi = MastervecFileObjectCount(HEADS[i] + ".fastb");
        String aligns_file;
        if ( !REF.Contains(".edit1") )
            aligns_file = HEADS[i] + ".vectrim.qltout." + REF.Before(".fasta");
        else
            aligns_file = workhead + "/" + HEADS[i].RevAfter("/")
                    + ".vectrim.qltout." + REF.Before(".fasta");
        LoadLookAligns(aligns_file, aligns, aligns_index, nreadsi);
        for ( int j = 0; j < nreadsi; j++ )
        {
            bvec const& read = bases_ref[j+readcount];
            if ( !aligns_index[j].solo() )
                continue;
            const look_align& la = aligns[aligns_index[j][0]];
            int tig = la.target_id, p1 = la.pos1(), p2 = la.pos2();
            int n = read.size();
            for ( int g = 0; g < la.a.Nblocks(); g++ )
            {
                if ( la.a.Gaps(g) > 0 )
                {
                    for ( int x = 0; x < la.a.Gaps(g); x++ )
                        ++calls[4][tig][p2 + x];
                    p2 += la.a.Gaps(g);
                }
                if ( la.a.Gaps(g) < 0 )
                {
                    for ( int x = 0; x < -la.a.Gaps(g); x++ )
                        ++calls[5][tig][p2];
                    p1 -= la.a.Gaps(g);
                }
                for ( int x = 0; x < la.a.Lengths(g); x++ )
                {
                    char base;
                    if ( !la.rc1 )
                        base = read[p1];
                    else
                        base = 3 - read[n - p1 - 1];
                    ++calls[base][tig][p2];
                    ++totCoverage;
                    ++p1;
                    ++p2;
                }
            }
        }
        readcount += nreadsi;
    }

    double maxCoverage = (3.*totCoverage)/ref.sumSizes();
    for ( size_t i = 0; i < ref.size(); i++ )
    {
        for ( unsigned int j = 0; j < ref[i].size(); j++ )
        {
            int M = calls[0][i][j], M2 = calls[1][i][j];
            if ( M+M2+calls[2][i][j]+calls[3][i][j] > maxCoverage )
                continue;

            int kmax = 0, k2max = 1;
            if ( M < M2 ) { using std::swap; swap(kmax,k2max); swap(M,M2); }
            for ( unsigned int k = 2; k < 6; ++k )
            {
                int val = calls[k][i][j];
                if ( val > M2 )
                {
                    if ( val > M )
                    {
                        k2max = kmax; M2 = M;
                        kmax = k; M = val;
                    }
                    else
                    {
                        k2max = k;
                        M2 = val;
                    }
                }
            }
            if ( kmax < 4 && M2 <= 3 && M >= 5 * M2 )
            {
                ref_fixed[i].Set(j, kmax);
                known[i].Set(j, True);
            }
        }
    }


    // pass 2
    readcount = 0;
    for ( int i = 0; i < HEADS.isize(); i++ )
    {
        vec<look_align> aligns;
        vec<vec<int> > aligns_index;
        int nreadsi = MastervecFileObjectCount(HEADS[i] + ".fastb");
        String aligns_file;
        if ( !REF.Contains(".edit1") )
        {
            aligns_file = HEADS[i] + ".vectrim.qltout." + REF.Before(
                    ".fasta");
        }
        else
        {
            aligns_file = workhead + "/" + HEADS[i].RevAfter("/")
                    + ".vectrim.qltout." + REF.Before(".fasta");
        }
        LoadLookAligns(aligns_file, aligns, aligns_index, nreadsi);
        for ( int j = 0; j < nreadsi; j++ )
        {
            bvec& read = bases_ref[j + readcount];
            if ( !aligns_index[j].solo() )
                continue;
            const look_align& la = aligns[aligns_index[j][0]];
            int tig = la.target_id, p1 = la.pos1(), p2 = la.pos2();
            int n = read.size();
            for ( int j = 0; j < la.a.Nblocks(); j++ )
            {
                if ( la.a.Gaps(j) > 0 )
                {
                    p2 += la.a.Gaps(j);
                }
                if ( la.a.Gaps(j) < 0 )
                {
                    p1 -= la.a.Gaps(j);
                }
                for ( int x = 0; x < la.a.Lengths(j); x++ )
                {
                    char base;
                    if ( !la.rc1 )
                        base = read[p1];
                    else
                        base = 3 - read[n - p1 - 1];

                    char refBase = ref_fixed[tig][p2];
                    if ( base != refBase && known[tig][p2] )
                    {
                        if ( !la.rc1 )
                            read.Set(p1, refBase);
                        else
                            read.Set(n - p1 - 1, 3 - refBase);
                    }
                    ++p1;
                    ++p2;
                }
            }
        }
        readcount += nreadsi;
    }
}

// Remove any edge that involves a source or a sink and does not
// contain the start or stop kmer of a reference contig.
void trimHyperKmerPath( HyperKmerPath& h, size_t nreads,
                        vec<big_tagged_rpint> const& unipathsdb,
                        vecKmerPath const& paths )
{
    // Find starts and stops of reference contigs.

    vec<int> starts, stops;
    for ( size_t i = nreads; i < paths.size( ); i++ )
    {    static vec<longlong> con;
         Contains( unipathsdb, paths[i].FirstSegment( ).Start( ), con );
         starts.push_back( unipathsdb[ con[0] ].ReadId( ) );
         Contains( unipathsdb, paths[i].LastSegment( ).Stop( ), con );
         stops.push_back( unipathsdb[ con[0] ].ReadId( ) );    }
    Sort(starts), Sort(stops);

    vec<int> sources, sinks;
    h.Sources(sources), h.Sinks(sinks);
    for ( int i = 0; i < sources.isize( ); i++ )
    {    int v = sources[i];
         if ( !h.Source(v) ) continue;
         for ( int j = h.From(v).isize( ) - 1; j >= 0; j-- )
         {    int e = h.EdgeObjectIndexByIndexFrom( v, j );
              if ( !BinMember( starts, e ) )
              {    sources.push_back( h.From(v)[j] );
                   h.DeleteEdgeFrom( v, j );    }    }    }
    for ( int i = 0; i < sinks.isize( ); i++ )
    {    int v = sinks[i];
         if ( !h.Sink(v) ) continue;
         for ( int j = h.To(v).isize( ) - 1; j >= 0; j-- )
         {    int e = h.EdgeObjectIndexByIndexTo( v, j );
              if ( !BinMember( stops, e ) )
              {    sinks.push_back( h.To(v)[j] );
                   h.DeleteEdgeTo( v, j );    }    }    }
    h.RemoveUnneededVertices( );
    h.RemoveDeadEdgeObjects( );
}

void trimReads( vecbvec& bases_ref, size_t nreads,
                vecKmerPath const& paths, vecKmerPath const& unipaths,
                unsigned int K )
{
    // Trim reads so that they contain only kmers in the unipaths.
    vec<longlong> ukmers;
    ukmers.reserve(unipaths.size()); // an underestimate, but might as well start with something

    for ( size_t j = 0; j < unipaths.size(); j++ )
    {
        KmerPathLoc l(unipaths[j].Begin());
        KmerPathLoc end(unipaths[j].End());
        ukmers.push_back(l.GetKmer());
        while ( l != end )
        {
            l = l + 1;
            ukmers.push_back(l.GetKmer());
        }
    }

    UniqueSort(ukmers);

    for ( size_t id = 0; id < nreads; id++ )
    {
        bvec& read = bases_ref[id];
        if ( read.size() < K )
        {
            read.clear();
            continue;
        }

        int count = 0;
        KmerPathLoc l(paths[id].Begin());
        KmerPathLoc end(paths[id].End());
        while ( BinMember(ukmers, l.GetKmer()) )
        {
            ++count;
            if ( l == end )
                break;
            l = l + 1;
        }
        read.resize(count + K - 1);
        if ( read.size() < K )
            read.clear();
    }
}

void deDupPlacements( size_t nreads, vec<placement>& places )
{
    vec<int> placements( nreads, 0 );
    for ( size_t i = 0; i < places.size( ); i++ )
         ++placements[ places[i].id ];
    vec<Bool> to_remove( places.size( ), False );
    for ( size_t i = 0; i < places.size( ); i++ )
         if ( placements[ places[i].id ] > 1 ) to_remove[i] = True;
    EraseIf( places, to_remove );
}

void alignToRef( ostream& out, String const& REFX, bool VERBOSE,
                 vecbvec const& bases_ref, size_t nreads,
                 vec<Bool>& placed1, vec<placement>& places )
{
    vec<look_align> aligns;
    PerfectLookup(12, bases_ref, REFX.Before(".fasta") + ".lookup", aligns,
                    FW_OR_RC, 0, nreads - 1);

    vec<vec<int> > aligns_index(nreads);
    for ( size_t i = 0; i < aligns.size(); i++ )
        aligns_index[aligns[i].query_id].push_back(i);

    // Phase 1': if a read was not placed in phase 1, but is uniquely placed on
    // the genome, place it.

    out << Date() << ": placing in phase 1'" << endl;
    size_t nplaced1p = 0;
    for ( size_t id = 0; id < nreads; id++ )
    {
        if ( !aligns_index[id].solo() || placed1[id] )
            continue;
        const look_align& la = aligns[aligns_index[id][0]];
        places.push(la.target_id, la.pos2(), id, la.Fw1());
        placed1[id] = True;
        if ( VERBOSE )
        {
            out << id << " placed by alignment at " << la.target_id << "."
                    << la.pos2() << (la.Fw1() ? "fw" : "rc") << "\n";
        }
        ++nplaced1p;
    }
    out << nplaced1p << " of " << nreads << " reads placed in phase1'" << endl;
}

void alignToRefPhase2( ostream& out, int pass, size_t nreads, unsigned int K,
                       unsigned int MAX_PLACEMENTS, bool VERBOSE,
                        double dinuke_entropy_floor,
                        vecKmerPath const& paths,
                        vecKmerPath const& pathsrc,
                        vec<big_tagged_rpint> const& pathsdb,
                        vec<big_tagged_rpint> const& gpathsdb,
                        vecbvec const& bases_ref,
                        vec<placement>& places,
                        vec<Bool>& placed1,
                        vec<placement>& places_all )
{
    vec<Bool> placed2(nreads, False);
    vec<vec<placement> > PLACES2(nreads);
    vec<Bool> overflowed(nreads, False);
    for ( size_t i = 0; i < places.size(); i++ )
    {
        if ( i % 100000 == 0 )
            PRINT3_TO( out, pass, i, places.size( ) );
        const placement& P = places[i];
        const KmerPath& r = (P.fw ? paths[P.id] : pathsrc[P.id]);
        int rpos = 0;
        for ( int j = 0; j < r.NSegments(); j++ )
        {
            static vec<longlong> con;
            Contains(gpathsdb, r.Segment(j), con);
            static vec<ho_interval> cov, uncov;
            cov.clear();
            for ( size_t u = 0; u < con.size(); u++ )
            {
                const big_tagged_rpint& t = gpathsdb[con[u]];
                int start = Max(0L, (t.Start() - r.Start(j)));
                int stop = Min(r.Length(j) - 1L, (t.Stop() - r.Start(j)));
                if ( start <= stop )
                    cov.push(start, stop + 1);
            }

            // Exclude low-entropy K-mers.

            static basevector b, c;
            b = bases_ref[P.id];
            if ( !P.fw )
                b.ReverseComplement();
            b.SetToSubOf(b, rpos, r.Length(j) + K - 1);
            for ( int u = 0; u < r.Length(j); u++ )
            {
                c.SetToSubOf(b, u, K);
                if ( DinukeEntropy(c) < dinuke_entropy_floor )
                    cov.push(u, u + 1);
            }
            rpos += r.Length(j);

            // Look for reads reaching off non-excluded K-mers.

            Uncovered(r.Length(j), cov, uncov);
            for ( size_t x = 0; x < uncov.size(); x++ )
            {
                KmerPathInterval I(r.Start(j) + uncov[x].Start(),
                        r.Start(j) + uncov[x].Stop() - 1);
                Contains(pathsdb, I, con);
                for ( size_t u = 0; u < con.size(); u++ )
                {
                    const big_tagged_rpint& t = pathsdb[con[u]];
                    int id = t.ReadId();
                    if ( static_cast<size_t>(id) >= nreads || placed1[id] )
                        continue;
                    if ( overflowed[id] )
                        continue;
                    const KmerPath& p = (t.Fw() ? paths[id] : pathsrc[id]);
                    placed2[id] = True;
                    int pp = t.PathPos();
                    int pos = P.pos + uncov[x].Start() + (int) (t.Start()
                            - I.Start());
                    for ( int y = 0; y < j; y++ )
                        pos += r.Length(y);
                    for ( int y = 0; y < pp; y++ )
                        pos -= p.Length(y);
                    placement P2(P.tig, pos, id, t.Fw());
                    if ( Member(PLACES2[id], P2) )
                        continue;
                    if ( PLACES2[id].size() == MAX_PLACEMENTS )
                    {
                        overflowed[id] = True;
                        PLACES2[id].clear();
                        placed2[id] = False;
                        continue;
                    }
                    if ( VERBOSE )
                    {
                        out << id << " placed via " << P.id << " at "
                                << P.tig << "." << pos
                                << (t.Fw() ? "fw" : "rc") << "\n";
                    }
                    PLACES2[id].push_back(P2);
                }
            }
        }
    }
    vec<placement> places2;
    for ( size_t i = 0; i < nreads; i++ )
        places2.append(PLACES2[i]);
    places_all.append(places2);
    out << Date() << ": ";
    PRINT_TO( out, Sum(placed2) );
    for ( size_t i = 0; i < nreads; i++ )
        if ( placed2[i] )
            placed1[i] = True;
    places = places2;
}

void packageMarks( size_t nreads, vec<placement>& places_all, String const& file )
{
    VecPlacementMarkVec marks(nreads);
    for ( size_t i = 0; i < places_all.size(); i++ )
    {
        const placement& p = places_all[i];
        marks[p.id].push_back( placement_mark(p.tig,p.fw,p.pos), 1.6f, 5u );
    }

    for ( size_t i = 0; i < nreads; i++ )
        sort_unique(marks[i]);

    marks.WriteAll(file);
}

vecKmerPath createUnipaths( ostream& out, unsigned int K, size_t nreads,
                            vecKmerPath const& paths,
                            vecKmerPath const& pathsrc,
                            vec<big_tagged_rpint> const& pathsdb )
{
    out << Date() << " Unipathing." << endl;
    vecKmerPath unipaths;
    vec<big_tagged_rpint> unipathsdb;
    Unipath( paths, pathsrc, pathsdb, unipaths, unipathsdb );
    HyperKmerPath h;
    if ( true )
    {
        digraph A;
        out << Date() << " Building adjacency graph." << endl;
        BuildUnipathAdjacencyGraph( paths, pathsrc, pathsdb, unipaths, unipathsdb, A );
        out << Date() << " Building hyperKmerPath." << endl;
        BuildUnipathAdjacencyHyperKmerPath( K, A, unipaths, h );
    }

    out << Date() << ' ' << h.EdgeObjectCount( ) << " edges" << endl;
    trimHyperKmerPath( h, nreads, unipathsdb, paths );

    unipaths.clear().reserve(h.EdgeObjectCount());
    for ( int i = 0; i < h.EdgeObjectCount( ); i++ )
         unipaths.push_back( h.EdgeObject(i) );
    CreateDatabase( unipaths, unipathsdb );
    out << Date() << ' ' << unipaths.size( ) << " unipaths" << endl;
    out << "N50 = " << h.EdgeN50( ) << endl;

    return unipaths;
}

int main( int argc, char *argv[] )
{
     RunTime( );

     BeginCommandArguments;
     CommandArgument_String(HEAD);
     CommandArgument_String(REF);
     CommandArgument_String_OrDefault( RUN, "" );
     CommandArgument_Int_OrDefault(TRUNC, 0);
     CommandArgument_String_OrDefault(DIR, "");
     CommandArgument_String_OrDefault(POST, "");
     CommandArgument_Int_OrDefault(ADD_PASSES, 100);
     CommandArgument_Int_OrDefault(MAX_PLACEMENTS, 10);
     CommandArgument_Bool_OrDefault(VERBOSE, False);
     CommandArgument_String_OrDefault(PIPELINE, "");
     EndCommandArguments;
     if ( PIPELINE == "" ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

     // Define directories, etc.

     String workdir, workhead;
     VAALSetup( RUN, HEAD, REF, DIR, POST, workdir, workhead, TRUNC, PIPELINE );
     cout << "writing to " << workhead + ".VAAL1.out\n" << endl;
     Ofstream( out, workhead + ".VAAL1.out" );
     command.PrintTheCommandPretty( out );

     // Load data.

     vecbasevector bases_ref, ref;
     String REFX;
     VAALLoad( workhead, HEAD, ".vectrim", TRUNC, REF, REFX, bases_ref,
          ref, PIPELINE );
     int nreads = bases_ref.size( );

     correctErrors( HEAD, TRUNC, PIPELINE, REF, workhead, ref, bases_ref );

     // Generate paths.

     bases_ref.append(ref.begin(),ref.end());
     out << Date( ) << ": collating reads" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXXXX
     vecKmerPath paths, pathsrc;
     vec<big_tagged_rpint> pathsdb;
     int K = 20;
     out << Date( ) << ": generating paths" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXXX
     ReadsToPathsCoreY( bases_ref, K, paths, pathsrc, pathsdb );
     paths.WriteAll( workhead + ".bases_ref.paths.20" );
     pathsrc.WriteAll( workhead + ".bases_ref.pathsrc.20" );
     //KmerBaseBrokerBig kbb( K, paths, pathsrc, pathsdb, bases_ref );
     out << Date( ) << ": paths loaded" << endl;

     // Create unipaths.
     if ( true )
     {
         vecKmerPath const& unipaths = createUnipaths(out,K,nreads,paths,pathsrc,pathsdb);
         out << Date( ) << ": trimming reads" << endl;
         trimReads(bases_ref,nreads,paths,unipaths,K);
     }

     // Write error-corrected, trimmed reads.  Rebuild paths.

     bases_ref.WriteRange( workhead + ".trimmed.fastb", 0, nreads );
     out << Date( ) << ": rebuilding paths" << endl;
     ReadsToPathsCoreY( bases_ref, K, paths, pathsrc, pathsdb );

     // Phase 1: for each K-mer that is unique in the reference, find all its
     // occurrences in the reads.

     out << Date( ) << ": placing reads - phase 1" << endl; // XXXXXXXXXXXXXXXXXXXX
     vec<placement> places;
     const double dinuke_entropy_floor = 1.7;
     AnchorReads( K, nreads, ref, paths, pathsrc, pathsdb, places,
          dinuke_entropy_floor );

     //TODO: done with ref here

     deDupPlacements( nreads, places );

     vec<Bool> placed1( nreads, False );
     for ( int i = 0; i < places.isize( ); i++ )
     {    placed1[ places[i].id ] = True;
          if (VERBOSE)
          {    out << places[i].id << " placed at " << places[i].tig << "."
                    << places[i].pos << ( places[i].fw ? "fw" : "rc" )
                    << "\n";    }    }

     vec<big_tagged_rpint> gpathsdb;
     CreateDatabase( vecKmerPath(paths.begin(nreads),paths.end()), gpathsdb );

     PRINT_TO( out, places.size( ) );
     out << Sum(placed1) << " of " << nreads << " reads placed in phase 1" << endl;
     out << Date( ) << ": phase 1 done" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

     // Align reads to the reference.

     out << Date( ) << ": aligning reads to reference" << endl;
     alignToRef( out, REFX, VERBOSE, bases_ref, nreads, placed1, places );
     BinaryWriter::writeFile( workhead + ".reference.placements", places );

     // Phase 2: for each K-mer in a placed read that is not in the reference, find
     // all the passing reads that contain it and are not already placed.  Add in
     // their inferred placements.  But do not place a read more than 10 times.

     vec<placement> places_all(places);
     for ( int pass = 1; pass <= ADD_PASSES; pass++ )
         alignToRefPhase2(out,pass,nreads,K,MAX_PLACEMENTS,VERBOSE,
                          dinuke_entropy_floor,paths,pathsrc,pathsdb,gpathsdb,
                          bases_ref,places,placed1,places_all);

     BinaryWriter::writeFile( workhead + ".reference.placements.all", places_all );

     out << Date( ) << ": packaging" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
     packageMarks( nreads, places_all, workhead+".marks");
}
