/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

// VAAL1.5.  Define safe intervals on the genome.  The idea is that they
// should be unique sequence for which the reference and sample genome agree.
// Currently they are computed in the following totally ad hoc way: they are the
// maximal intervals of size >= 60 whose constituent 20-mers are unique in the
// reference and occur at least 5 times in the reads, and which are not adjacent in
// the reads to an alternate kmer (occurring at least 5 times), with the outer 20
// bases on each side trimmed off.
//
// Modified subsequently.

#include "Basevector.h"
#include "Bitvector.h"
#include "FeudalMimic.h"
#include "MainTools.h"
#include "lookup/LookAlign.h"
#include "lookup/PerfectLookup.h"
#include "paths/KmerPath.h"
#include "paths/ReadsToPathsCoreX.h"
#include "polymorphism/VAALTools.h"
#include "solexa/SolexaTools.h"

int main( int argc, char *argv[] )
{
     RunTime( );

     BeginCommandArguments;
     CommandArgument_String(HEAD);
     CommandArgument_String(REF);
     CommandArgument_String_OrDefault(TRUE_REF, "");
     CommandArgument_String_OrDefault( RUN, "" );
     CommandArgument_Int_OrDefault(TRUNC, 0);
     CommandArgument_String_OrDefault(DIR, "");
     CommandArgument_String_OrDefault(PRE, "");
     CommandArgument_String_OrDefault(POST, "");
     CommandArgument_String_OrDefault(PIPELINE, "");
     CommandArgument_Bool_OrDefault(ASSESS, False);
     CommandArgument_Bool_OrDefault(WRITE, True);
     EndCommandArguments;
     if ( PIPELINE == "" ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

     // Define directories, etc.

     String workdir, workhead;
     VAALSetup( RUN, HEAD, REF, DIR, POST, workdir, workhead, TRUNC,
          PIPELINE );

     // Sanity check.

     if (ASSESS) ForceAssert( TRUE_REF != "" );

     // Load data.

     vecbasevector bases, ref;
     String REFX;
     VAALLoad( workhead, HEAD, ".vectrim", TRUNC, REF, REFX, bases,
          ref, PIPELINE );
     int nreads = bases.size( );

     // Load paths for reads and reference.

     vecKmerPath paths, pathsrc;
     vec<big_tagged_rpint> pathsdb;
     paths.ReadAll( workhead + ".bases_ref.paths.20" );
     pathsrc.ReadAll( workhead + ".bases_ref.pathsrc.20" );
     CreateDatabase( paths, pathsrc, pathsdb );

     // Find safe spots.  Note that what we compute here (safe0) only marks the
     // start points of K-mers.

     vecbitvector safe0, safe;
     Mimic( ref, safe0 );
     Mimic( ref, safe );
     for ( size_t tig = 0; tig < ref.size( ); tig++ )
     {    const KmerPath& R = paths[ bases.size( ) + tig ];
          KmerPathLoc l = R.Begin( );
          int pos = 0;
          while(1)
          {    longlong x = l.GetKmer( );
               static vec<longlong> con;
               Contains( pathsdb, x, con );
               int readcount = 0, refcount = 0;
               for ( int j = 0; j < con.isize( ); j++ )
               {    const big_tagged_rpint& t = pathsdb[ con[j] ];
                    if ( static_cast<size_t>(t.ReadId( )) < bases.size( ) ) ++readcount;
                    else ++refcount;    }

               // Look for branches.

               Bool branch = False;
               const int min_branch = 3;
               if ( l != R.Begin( ) && l != R.End( ) )
               {    longlong next = (l+1).GetKmer( ), back = (l-1).GetKmer( );
                    static vec<longlong> nexts, backs;
                    nexts.clear( ), backs.clear( );
                    for ( int z = 0; z < con.isize( ); z++ )
                    {    const big_tagged_rpint& t = pathsdb[ con[z] ];
                         size_t id = t.ReadId( );
                         if ( id >= bases.size( ) ) continue;
                         const KmerPath& p = ( t.Fw( ) ? paths[id] : pathsrc[id] );
                         int pp = t.PathPos( );
                         if ( x+1 <= t.Stop( ) ) nexts.push_back(x+1);
                         else if ( pp+1 < p.NSegments( ) )
                         {    nexts.push_back( p.Start(pp+1) );    }    }
                    Sort(nexts);
                    for ( int i = 0; i < nexts.isize( ); i++ )
                    {    int j = nexts.NextDiff(i);
                         if ( j - i >= min_branch && nexts[i] != next )
                              branch = True;
                         i = j - 1;    }
                    for ( int z = 0; z < con.isize( ); z++ )
                    {    const big_tagged_rpint& t = pathsdb[ con[z] ];
                         size_t id = t.ReadId( );
                         if ( id >= bases.size( ) ) continue;
                         const KmerPath& p = ( t.Fw( ) ? paths[id] : pathsrc[id] );
                         int pp = t.PathPos( );
                         if ( x-1 <= t.Start( ) ) backs.push_back(x-1);
                         else if ( pp > 0 ) backs.push_back( p.Stop(pp-1) );    }
                    Sort(backs);
                    for ( int i = 0; i < backs.isize( ); i++ )
                    {    int j = backs.NextDiff(i);
                         if ( j - i >= min_branch && backs[i] != back )
                              branch = True;
                         i = j - 1;    }    }

               // Record result.

               if ( refcount == 1 && readcount >= 5 && !branch )
                    safe0[tig].Set( pos, True );
               if ( l == R.End( ) ) break;
               l = l + 1;
               pos++;    }

          // Phase I: form strong safe spots.

          vec<ho_interval> spots;
          for ( unsigned int p = 0; p < safe[tig].size( ); p++ )
          {    if ( !safe0[tig][p] ) continue;
               unsigned int q;
               for ( q = p + 1; q < safe[tig].size( ); q++ )
                    if ( !safe0[tig][q] ) break;
               if ( q - p >= 40 )
               {    for ( unsigned int j = p + 20; j < q; j++ )
                         safe[tig].Set( j, True );
                    spots.push( p + 20, q );    }
               p = q - 1;    }

          // Phase II: if two consecutive strong safe spots are separated by more
          // than 300 bp, find the safe spot closest to the midpoint, and save it.
          // Iterate.

          while(1)
          {    int nspots = spots.size( );
               for ( int i = 1; i < nspots; i++ )
               {    int from = spots[i-1].Stop( ), to = spots[i].Start( );
                    if ( to - from <= 300 ) continue;
                    int mid = (from+to)/2;
                    for ( int j = 0; j < mid-from; j++ )
                    {    if ( safe0[tig][mid-j] )
                         {    int k;
                              for ( k = mid-j-1; k >= 0; k-- )
                                   if ( !safe0[tig][k] ) break;
                              spots.push( k+1, mid-j+1+20 );
                              for ( int u = spots.back( ).Start( );
                                   u < spots.back( ).Stop( ); u++ )
                              {    safe[tig].Set( u, True );    }
                              break;    }
                         if ( safe0[tig][mid+j] )
                         {    int k;
                              for ( k = mid+j+1; k < to; k++ )
                                   if ( !safe0[tig][k] ) break;
                              spots.push( mid+j, k+20 );
                              for ( int u = spots.back( ).Start( );
                                   u < spots.back( ).Stop( ); u++ )
                              {    safe[tig].Set( u, True );    }
                              break;    }    }    }
               if ( nspots == spots.isize( ) ) break;
               Sort(spots);    }    }

     // Assess safe intervals.

     if (ASSESS)
     {    vecbasevector safetigs;
          vec<look_align> aligns;
          vec< vec<int> > aligns_index;
          for ( int pass = 1; pass <= 2; pass++ )
          {    if ( pass == 2 )
               {    if ( !TRUE_REF.Contains( ".fasta", -1 ) ) TRUE_REF += ".fasta";
                    String TRUE_REFX;
                    if ( TRUE_REF != "" )
                    {    vec<String> HEADS;
                         ExpandHead( HEAD, HEADS, TRUNC, PIPELINE );
                         TRUE_REFX = TRUE_REF;
                         String run = HEADS[0].RevBefore( "/" );
                         if ( IsRegularFile( run + "/" + TRUE_REF ) )
                              TRUE_REFX = run + "/" + TRUE_REF;
                         else if ( run.Contains( "/" ) )
                         {    run = run.RevBefore( "/" );
                              if ( IsRegularFile( run + "/" + TRUE_REF ) )
                                   TRUE_REFX = run + "/" + TRUE_REF;    }    }
                    String lookup = TRUE_REFX.Before( ".fasta" ) + ".lookup";
                    PerfectLookup( 12, safetigs, lookup, aligns, FW_OR_RC );
                    aligns_index.resize( safetigs.size( ) );
                    for ( int i = 0; i < aligns.isize( ); i++ )
                         aligns_index[ aligns[i].query_id ].push_back(i);    }
               int count = 0;
               for ( size_t t = 0; t < safe.size( ); t++ )
               {    for ( unsigned int i = 0; i < safe[t].size( ); i++ )
                    {    if ( !safe[t][i] ) continue;
                         unsigned int j;
                         for ( j = i + 1; j < safe[t].size( ); j++ )
                              if ( !safe[t][j] ) break;
                         static basevector s;
                         if ( pass == 1 )
                         {    s.SetToSubOf( ref[t], i, j - i );
                              safetigs.push_back_reserve(s);    }
                         if ( pass == 2 )
                         {    cout << t << "." << i << "-" << j;
                              if ( aligns_index[count].empty( ) ) cout << " WRONG!";
                              if ( aligns_index[count].size( ) > 1 )
                                   cout << " MULTIPLE IN SAMPLE!";
                              cout << "\n";
                              ++count;    }
                         i = j - 1;    }    }    }    }

     // Write results.

     if (WRITE) safe.WriteAll( workhead + ".safespots" );    }
