/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

// VAAL2.  Assemble unpaired reads, using a related strain for
// assistance.  You have to have run VAAL1 and VAAL1.5 first.
//
// Ultimate purpose: Identify polymorphisms in reads as compared to a given
// reference.  In progress.
//
// START=tig.pos = starting position for assembly.  If unspecified, assemble the
// entire genome.
//
// OUT = fastb output file; not required.  If both OUT and START are unspecified,
// used canonical output file HEAD.assembly0.fastb.
//
// See VAAL1.cc for general info.

#include "Basevector.h"
#include "Equiv.h"
#include "MainTools.h"
#include "ParseSet.h"
#include "STLExtensions.h"
#include "lookup/LookAlign.h"
#include "lookup/PerfectLookup.h"
#include "math/Functions.h"
#include "math/HoInterval.h"
#include "paths/AlignHyperKmerPath.h"
#include "paths/AlignSeqsToHyper.h"
#include "paths/KmerBaseBroker.h"
#include "paths/KmerPath.h"
#include "paths/ReadsToPathsCoreX.h"
#include "paths/SeqOnHyper.h"
#include "paths/Unipath.h"
#include "polymorphism/VAALTools.h"
#include "solexa/SolexaTools.h"
#include "feudal/BinaryStream.h"
#include "feudal/VirtualMasterVec.h"

using namespace std;

// Trim reads so that they contain only kmers in the unipaths.

void TrimReads( const vecbasevector& reads, vecbasevector& reads2, const int KX,
     const vecKmerPath& unipaths, const vec<Bool>& fw, const vecKmerPath& paths )
{    reads2 = reads;
     vec<longlong> ukmers;
     for ( size_t j = 0; j < unipaths.size( ); j++ )
     {    KmerPathLoc l = unipaths[j].Begin( );
          while(1)
          {    ukmers.push_back( l.GetKmer( ) );
               if ( l == unipaths[j].End( ) ) break;
               l = l + 1;    }    }
     UniqueSort(ukmers);
     for ( int id = 0; id < fw.isize( ); id++ )
     {    if ( reads[id].isize( ) < KX ) continue;
          int count = 0;
          KmerPathLoc l = ( fw[id] ? paths[id].Begin( ) : paths[id].End( ) );
          while(1)
          {    if ( !BinMember( ukmers, l.GetKmer( ) ) ) break;
               ++count;
               if ( l == ( fw[id] ? paths[id].End( ) : paths[id].Begin( ) ) )
                    break;
               if ( fw[id] ) l = l + 1;
               else l = l - 1;    }
          if ( !fw[id] ) reads2[id].ReverseComplement( );
          reads2[id].resize( count + KX - 1 );
          if ( !fw[id] ) reads2[id].ReverseComplement( );    }    }

// Remove any edge that involves a source or a sink and does not contain
// the start or stop kmer of the reference chunk.

void TrimHair( HyperKmerPath& h, const KmerBaseBroker& kbb,
     const vecKmerPath& paths, const vecKmerPath& pathsrc,
     const vec<tagged_rpint>& pathsdb, const int KX, const basevector& xxx,
     vecKmerPath& unipaths, vec<tagged_rpint>& unipathsdb,
     const longlong x1, const longlong x2, const Bool ASSESS, ostream& out )
{
     while(1)
     {    // Find start and stop kmers of reference chunk.

          static vec<longlong> con;
          Contains( unipathsdb, x1, con );
          if ( con.empty( ) )
          {    out << "Weird, can't find x1." << endl;
               out << "Abandoning haircut." << endl;
               return;    }
          int u1 = unipathsdb[ con[0] ].ReadId( );
          Contains( unipathsdb, x2, con );
          if ( con.empty( ) )
          {    out << "Weird, can't find x2." << endl;
               out << "Abandoning haircut." << endl;
               return;    }
          int u2 = unipathsdb[ con[0] ].ReadId( );

          if (ASSESS) AssessUnipaths( h, kbb, KX, xxx, out );

          // TODO: potentially dangerous truncation of index by fu
          vec< pair<int,int> > fu;
          for ( size_t id = 0; id < unipaths.size( ); id++ )
          {    int f = xxx.Find( kbb.Seq( unipaths[id] ) );
               if ( f < xxx.isize( ) ) fu.push( f, id );    }
          Sort(fu);

          // Remove edges.

          vec<Bool> to_remove( h.EdgeObjectCount( ), False );
          for ( int v = 0; v < h.N( ); v++ )
          {    if ( h.Source(v) )
               {    for ( int j = 0; j < h.From(v).isize( ); j++ )
                    {    int e = h.EdgeObjectIndexByIndexFrom( v, j );
                         if ( e == u1 || e == u2 ) continue;
                         // out << "removing " << e << endl; // XXXXXXXXXXXXXXXXXXX
                         for ( int z = 0; z < fu.isize( ); z++ )
                         {    if ( e == fu[z].second )
                                   out << "DELETING GOOD UNIPATH!\n";    }
                         to_remove[e] = True;    }    }
               if ( h.Sink(v) )
               {    for ( int j = 0; j < h.To(v).isize( ); j++ )
                    {    int e = h.EdgeObjectIndexByIndexTo( v, j );
                         if ( e == u1 || e == u2 ) continue;
                         // out << "removing " << e << endl; // XXXXXXXXXXXXXXXXXXX
                         for ( int z = 0; z < fu.isize( ); z++ )
                         {    if ( e == fu[z].second )
                                   out << "DELETING GOOD UNIPATH!\n";    }
                         to_remove[e] = True;    }    }    }
          if ( Sum(to_remove) == 0 ) break;
          unipaths.EraseIf(to_remove);
          ClipToStartStop( unipaths, x1, x2 );

          Rebuild( unipaths, unipathsdb, h, KX, paths, pathsrc, pathsdb );    }    }

// Placement mark that also keeps the read id

class placement_mark_plus {

     public:

     placement_mark_plus( ) { }
     placement_mark_plus( const int id, const placement_mark& m ) : id(id), m(m) { }

     int id;
     placement_mark m;

     friend Bool operator<( const placement_mark_plus& M1,
          const placement_mark_plus& M2 )
     {    if ( M1.m.Tig( ) < M2.m.Tig( ) ) return True;
          if ( M1.m.Tig( ) > M2.m.Tig( ) ) return False;
          if ( M1.m.Pos( ) < M2.m.Pos( ) ) return True;
          return False;    }

};

basevector AssembleWindow(
     const int TIG, const int Start, const int STOP,
     const vecbasevector& ref, const String TRUE_REFX, const String REFX,
     const vec<placement_mark_plus>& marks_plus,
     const vecbasevector& bases,
     const Bool PRINT_IDS, const int DUMP_NHOOD_READS,
     ostream& out, const int KX, int& min_over, Bool ASSESS, const int EXTRA,
     vecbasevector& extra_reads, const String& workhead )
{
     out << "\nASSEMBLING " << TIG << "." << Start << "-" << STOP
          << " using KX = " << KX << endl << endl;

     // Find [Start,STOP) on truth, as basevector xxx.

     basevector xxx;
     if (ASSESS)
     {    vecbasevector ttt(1);
          ttt[0].SetToSubOf( ref[TIG], Start, 20 );
          String lookup = TRUE_REFX.Before( ".fasta" ) + ".lookup";
          vec<look_align> aligns;
          PerfectLookup( 12, ttt, lookup, aligns, FW_OR_RC );
          if ( !aligns.solo( ) )
          {    out << "Can't place start uniquely on truth.\n";
               out << "Turning assessment off.\n";
               ASSESS = False;    }
          if (ASSESS)
          {    look_align start_a = aligns[0];
               int tig = aligns[0].target_id;
               Bool Fw1 = aligns[0].Fw1( );
               ttt[0].SetToSubOf( ref[TIG], STOP-20, 20 );
               PerfectLookup( 12, ttt, lookup, aligns, FW_OR_RC );
               if ( !aligns.solo( ) )
               {    out << "Can't place stop uniquely on truth.\n";
                    out << "Turning assessment off.\n";
                    ASSESS = False;    }
               if (ASSESS)
               {    look_align stop_a = aligns[0];
                    PRINT2_TO( out, tig, aligns[0].target_id ); // XXX
                    PRINT2_TO( out, int(Fw1), int( aligns[0].Fw1( ) ) ); // XXX
                    PRINT2_TO( out, start_a.pos2( ), start_a.Pos2( ) ); // XXX
                    PRINT2_TO( out, stop_a.pos2( ), stop_a.Pos2( ) ); // XXX
                    if ( tig != aligns[0].target_id || Fw1 != aligns[0].Fw1( )
                         || ( Fw1 && !( start_a.pos2( ) < stop_a.Pos2( ) ) )
                         || ( !Fw1 && !( stop_a.pos2( ) < start_a.Pos2( ) ) ) )
                    {    out << "Can't place start and stop consistently "
                              << "on truth.\n";
                         out << "Turning assessment off.\n";
                         ASSESS = False;    }
                    if (ASSESS)
                    {    int tstart = ( Fw1 ? start_a.pos2( ) : stop_a.pos2( ) );
                         int tstop = ( Fw1 ? stop_a.Pos2( ) : start_a.Pos2( ) );
                         vecbasevector genome(
                              TRUE_REFX.Before( ".fasta" ) + ".fastb" );
                         xxx.SetToSubOf( genome[tig], tstart, tstop - tstart );
                         if ( !Fw1 ) xxx.ReverseComplement( );    }    }    }    }

     // Define wingsize.

     // Find all the reads falling in the interval:
     // from all passed placement_mark_plus'es, select only those
     // that are on requested contig TIG and are completely (i.e. without
     // overhang) between positions Start - EXTRA and STOP + EXTRA

     vec<int> ids; // will keep ids of reads that we select for further work
     vec<Bool> fw; // fw/rc flag; synchronized with 'ids' list (above)
     vec<placement_mark> Marks; // placement marks (positions) of the selected
                                // reads on the reference - synchronized with ids

     // a "tag" - placement mark located exactly at the start of the
     // interval we are interested in:

     placement_mark_plus m( 0, placement_mark( TIG, True, Start - EXTRA ) );

     // find index of the first mark >= m:

     int start = upper_bound( marks_plus.begin( ), marks_plus.end( ), m )
          - marks_plus.begin( );
     if ( start > 0 ) --start; // roll back to the last mark < m

     for ( int i = start; i < marks_plus.isize( ); i++ )
     {    const placement_mark_plus& M = marks_plus[i];
          if ( M.m.Tig( ) > TIG ) break;
          if ( (int) M.m.Pos( ) >= STOP + EXTRA ) break;
          if ( M.m.Tig( ) != TIG ) continue;
          if ( (int) M.m.Pos( ) < Start - EXTRA ) continue;
          int id = M.id;
          if ( (int) M.m.Pos( ) + bases[id].isize( ) > STOP + EXTRA ) continue;
          if (PRINT_IDS) out << id << "\n";
          if (DUMP_NHOOD_READS >= 1)
          {    out << ">" << ids.size( ) << " (" << id << "->"
	           << TIG << "." << M.m.Pos( ) << "-"
	           << M.m.Pos( ) + bases[id].isize( )
	           << ( M.m.Fw1( ) ? "fw" : "rc" ) << ")\n";
               if ( DUMP_NHOOD_READS == 1 )
                    out << bases[id].ToString( ) << "\n";    }
          ids.push_back(id);
          fw.push_back( M.m.Fw1( ) );
          Marks.push_back(M.m);    }

     // Build paths.

     vecbasevector reads;
     for ( int i = 0; i < ids.isize( ); i++ )
     {    int id = ids[i];
          if ( fw[i] ) reads.push_back_reserve( bases[id] );
          else
          {    static basevector r;
               r = bases[id];
               r.ReverseComplement( );
               reads.push_back_reserve(r);    }    }
     reads.Append(extra_reads);

     // Form the reads plus the reference chunk.

     basevector refchunk;
     refchunk.SetToSubOf( ref[TIG], Start, STOP-Start );
     out << "refchunk = " << refchunk.ToString( ) << "\n\n";
     vecbasevector readsplus(reads);
     readsplus.push_back_reserve(refchunk);

     // Build the unipath graph.

     vecKmerPath paths, pathsrc;
     vec<tagged_rpint> pathsdb;
     ReadsToPathsCoreY( readsplus, KX, paths );
     CreateDatabase( paths, pathsdb );
     KmerBaseBroker kbb( KX, paths, pathsrc, pathsdb, readsplus );

     // Workaround for bug.

     for ( size_t i = 0; i < paths.size( ); i++ )
     {    if ( !kbb.ToSequenceSafe(paths[i]) )
          {    cout << "Observe bug in paths, giving up\n";
               return basevector( );    }    }

     vecKmerPath unipaths;
     vec<tagged_rpint> unipathsdb;
     Unipath( paths, pathsrc, pathsdb, unipaths, unipathsdb );
     digraph A;
     BuildUnipathAdjacencyGraph( paths, pathsrc, pathsdb, unipaths,
          unipathsdb, A );
     HyperKmerPath h;
     BuildUnipathAdjacencyHyperKmerPath( KX, A, unipaths, h );

     // Trim reads so that they contain only kmers in the unipaths.

     vecbasevector reads2;
     TrimReads( reads, reads2, KX, unipaths, fw, paths );

     // Remove any edge that involves a source or a sink and does not contain
     // the start or stop kmer of the reference chunk.

     longlong x1 = paths[ reads.size( ) ].FirstSegment( ).Start( );
     longlong x2 = paths[ reads.size( ) ].LastSegment( ).Stop( );
     TrimHair( h, kbb, paths, pathsrc, pathsdb, KX, xxx, unipaths, unipathsdb,
          x1, x2, ASSESS, out );
     if (ASSESS) AssessUnipaths( h, kbb, KX, xxx, out );

     // At each branch, let the reads vote.  Strongly outvoted edges are
     // removed.

     for ( int pass = 1; pass <= 2; pass++ )
     {
          // Align the reads to the graph.

          HyperBasevector hb( h, kbb );
          vec<SeqOnHyper> saligns;
          AlignSeqsToHyper( hb, reads2, saligns );

          vec<Bool> remove( h.EdgeObjectCount( ), False );

          for ( int v = 0; v < h.N( ); v++ )
          {    if ( h.From(v).size( ) > 1 )
               {    vec<int> votes( h.From(v).size( ), 0 );
                    for ( int i = 0; i < saligns.isize( ); i++ )
                    {    const SeqOnHyper& ap = saligns[i];
                         for ( int j = 1; j < ap.N( ); j++ )
                         {    int p = Position( h.FromEdgeObj(v), ap.Id2(j) );
                              if ( p < 0 ) continue;
                              ++votes[p];    }    }
                    out << "\nexiting vertex " << v << ":\n";
                    for ( int j = 0; j < h.From(v).isize( ); j++ )
                    {    int e = h.EdgeObjectIndexByIndexFrom( v, j );
                         out << votes[j] << " votes for " << BaseAlpha(e) << "\n";
                         for ( int k = 0; k < votes.isize( ); k++ )
                         {    if ( votes[j] >= 3 && votes[j] >= 3 * votes[k] )
                              {    if ( votes[k] > 3 ) continue;
                                   int ealt = h.EdgeObjectIndexByIndexFrom( v, k );
                                   if ( remove[ealt] ) continue;
                                   out << "KILL edge " << BaseAlpha(ealt)
                                        << "\n";
                                   remove[ealt] = True;    }    }    }    }
               if ( h.To(v).size( ) > 1 )
               {    vec<int> votes( h.To(v).size( ), 0 );
                    for ( int i = 0; i < saligns.isize( ); i++ )
                    {    const SeqOnHyper& ap = saligns[i];
                         for ( int j = 0; j < ap.N( ) - 1; j++ )
                         {    int u = ap.Id2(j);
                              int p = Position( h.ToEdgeObj(v), u );
                              if ( p < 0 ) continue;
                              ++votes[p];    }    }
                    out << "\nentering vertex " << v << ":\n";
                    for ( int j = 0; j < h.To(v).isize( ); j++ )
                    {    int e = h.EdgeObjectIndexByIndexTo( v, j );
                         out << votes[j] << " votes for " << BaseAlpha(e) << "\n";
                         for ( int k = 0; k < votes.isize( ); k++ )
                         {    if ( votes[j] >= 3 && votes[j] >= 3 * votes[k] )
                              {    if ( votes[k] > 3 ) continue;
                                   int ealt = h.EdgeObjectIndexByIndexTo( v, k );
                                   if ( remove[ealt] ) continue;
                                   out << "KILL edge " << BaseAlpha(ealt)
                                        << "\n";
                                   remove[ealt] = True;    }    }    }    }    }


          for ( int v = 0; v < h.N( ); v++ )
          {    out << "exiting vertex " << v << ":\n";
               for ( int j1 = 0; j1 < h.From(v).isize( ); j1++ )
               {    vec<int> e(2), l(2), s(2);
                    e[0] = h.EdgeObjectIndexByIndexFrom( v, j1 );
                    l[0] = h.EdgeLength(e[0]);
                    for ( int j2 = 0; j2 < h.From(v).isize( ); j2++ )
                    {    if ( j2 == j1 ) continue;
                         e[1] = h.EdgeObjectIndexByIndexFrom( v, j2 );
                         l[1] = h.EdgeLength(e[1]);
                         int minkmers = Min( l[0], l[1] );
                         vec< vec<int> > support(2);
                         for ( int i = 0; i < saligns.isize( ); i++ )
                         {    const SeqOnHyper& ap = saligns[i];
                              for ( int j = 0; j < ap.N( ); j++ )
                              {    int x = Position( e, ap.Id2(j) );
                                   if ( x < 0 ) continue;
                                   if ( ap.pos2(j) >= minkmers ) continue;
                                   support[x].push_back( ap.Id1( ) );    }    }
                         for ( int x = 0; x < 2; x++ )
                              UniqueSort( support[x] );
                         s[0] = support[0].size( ), s[1] = support[1].size( );
                         out << BaseAlpha(e[0]) << " = " << s[0] << ", vs "
                              << BaseAlpha(e[1]) << " = " << s[1];
                         if ( s[0] >= 3 && s[0] >= 3 * s[1] )
                         {    out << " --> KILLING " << BaseAlpha(e[1])
                                   << " = "
                                   << kbb.Seq( unipaths[ e[1] ] ).ToString( );
                              remove[ e[1] ] = True;    }
                         out << "\n";    }    }
               out << "entering vertex " << v << ":\n";
               for ( int j1 = 0; j1 < h.To(v).isize( ); j1++ )
               {    vec<int> e(2), l(2), s(2);
                    e[0] = h.EdgeObjectIndexByIndexTo( v, j1 );
                    l[0] = h.EdgeLength(e[0]);
                    for ( int j2 = 0; j2 < h.To(v).isize( ); j2++ )
                    {    if ( j2 == j1 ) continue;
                         e[1] = h.EdgeObjectIndexByIndexTo( v, j2 );
                         l[1] = h.EdgeLength(e[1]);
                         int minkmers = Min( l[0], l[1] );
                         vec< vec<int> > support(2);
                         for ( int i = 0; i < saligns.isize( ); i++ )
                         {    const SeqOnHyper& ap = saligns[i];
                              for ( int j = 0; j < ap.N( ); j++ )
                              {    int x = Position( e, ap.Id2(j) );
                                   if ( x < 0 ) continue;
                                   if ( l[x] + KX - 1 - ap.Pos2(j) >= minkmers )
                                        continue;
                                   support[x].push_back( ap.Id1( ) );    }    }
                         for ( int x = 0; x < 2; x++ )
                              UniqueSort( support[x] );
                         s[0] = support[0].size( ), s[1] = support[1].size( );
                         out << BaseAlpha(e[0]) << " = " << s[0] << ", vs "
                              << BaseAlpha(e[1]) << " = " << s[1];
                         if ( s[0] >= 3 && s[0] >= 3 * s[1] )
                         {    out << " --> KILLING " << BaseAlpha(e[1]);
                              remove[ e[1] ] = True;    }
                         out << "\n";    }    }    }

          unipaths.EraseIf(remove);
          ClipToStartStop( unipaths, x1, x2 );
          Rebuild( unipaths, unipathsdb, h, KX, paths, pathsrc, pathsdb );

          // Delete isolated loops.

          vec<Bool> isolated( h.EdgeObjectCount( ), False );
          for ( int v = 0; v < h.N( ); v++ )
          {    if ( h.From(v).solo( ) && h.To(v).solo( ) )
               {    int e1 = h.EdgeObjectIndexByIndexFrom( v, 0 );
                    int e2 = h.EdgeObjectIndexByIndexTo( v, 0 );
                    if ( e1 == e2 ) isolated[e1] = True;    }    }
          unipaths.EraseIf(isolated);
          Rebuild( unipaths, unipathsdb, h, KX, paths, pathsrc, pathsdb );
          if (ASSESS) AssessUnipaths( h, kbb, KX, xxx, out );

          // Print the graph.

          // Ofstream( dot2, "yyy.dot" );
          // h.PrintSummaryDOT0w( dot2, False, False, True );

          if (ASSESS)
          {    for ( int pass = 1; pass <= 2; pass++ )
               {    String GENOME;
                    if ( pass == 1 )
                    {    if ( TRUE_REFX.empty() ) continue;
                         GENOME = TRUE_REFX.Before( ".fasta" );    }
                    else GENOME = REFX.Before( ".fasta" );
                    out << "\nALIGNMENT TO ";
                    if ( pass == 1 ) out << "SAMPLE GENOME:\n";
                    else out << "RELATED REFERENCE:\n";
                    {    static vec<look_align> aligns;
                         static vec< vec<int> > aligns_index;
                         static vec<TrustedPath> trusted_paths;
                         AlignHyperKmerPath( h, &kbb, GENOME, workhead, aligns,
                              aligns_index );
                         // FilterAligns( h, aligns, aligns_index, trusted_paths );
                         vecbasevector genome( GENOME + ".fastb" );
                         PrintAlignedHyperKmerPath( out, h, &kbb, genome, aligns,
                              aligns_index, False, &trusted_paths );    }    }    }

          out << "\nREBUILT GRAPH\n";
          h.PrintSummaryPlus(out);
          out << "\n";
          for ( int i = 0; i < h.EdgeObjectCount( ); i++ )
          {    out << ">[" << BaseAlpha(i) << "]\n"
                    << kbb.Seq( h.EdgeObject(i) ).ToString( ) << "\n";    }    }

     // Again remove any edge that involves a source or a sink and does not contain
     // the start or stop kmer of the reference chunk.

     TrimHair( h, kbb, paths, pathsrc, pathsdb, KX, xxx, unipaths, unipathsdb,
          x1, x2, ASSESS, out );
     if (ASSESS) AssessUnipaths( h, kbb, KX, xxx, out );

     // Append to extra_reads.

     for ( size_t i = 0; i < unipaths.size( ); i++ )
          extra_reads.push_back_reserve( kbb.Seq( unipaths[i] ) );

     // Decide if we're done.

     if ( unipaths.size( ) == 1 && unipaths[0].FirstSegment( ).Start( ) == x1
          && unipaths[0].LastSegment( ).Stop( ) == x2 )
     {    out << "\nACCEPTED!" << endl;

          basevector answer = kbb.Seq( unipaths[0] );

          // See how well covered the answer is.  We fix K to allow comparisons
          // between what one would get with different values of KX.

          const int K = 20;
          basevector answer_plus = Cat( basevector( ref[TIG], Start - K, K ),
               kbb.Seq( unipaths[0] ) );
          if( STOP + K <= ref[TIG].isize( ) )
               answer_plus = Cat( answer_plus, basevector( ref[TIG], STOP, K ) );
          vecbasevector readsplus(reads);
          readsplus.push_back_reserve(answer_plus);
          vecKmerPath paths, pathsrc;
          vec<tagged_rpint> pathsdb;
          ReadsToPathsCoreY( readsplus, K, paths );
          CreateDatabase( paths, pathsdb );

          // Workaround for bug.

          KmerBaseBroker kbbx( K, paths, pathsrc, pathsdb, readsplus );
          for ( size_t i = 0; i < paths.size( ); i++ )
          {    if ( !kbbx.ToSequenceSafe(paths[i]) )
               {    cout << "Observe Bug in paths, giving up\n";
                    return basevector( );    }    }

          vecKmerPath unipaths;
          unipaths.push_back_reserve( paths[ reads.size( ) ] );
          vec<basevector> x;
          x.push_back(answer_plus);
          HyperBasevector hb( K, x );
          TrimReads( reads, reads2, K, unipaths, fw, paths );
          vec<SeqOnHyper> saligns;
          AlignSeqsToHyper( hb, reads2, saligns );
          vec<ho_interval> cov;
          for ( int i = 0; i < saligns.isize( ); i++ )
          {    const SeqOnHyper& ap = saligns[i];
               cov.push( ap.pos2(0), ap.Pos2(0) );    }
          Sort(cov);
          Bool cov_verbose = False;
          if (cov_verbose) PRINT_TO( out, answer_plus.size( ) );
          if ( cov.empty( ) ) out << "\nnot covered at all\n";
          else
          {    int last = cov[0].Stop( );
               min_over = 1000000;
               if (cov_verbose) out << cov[0] << endl;
               for ( int j = 1; j < cov.isize( ); j++ )
               {    if ( cov[j].Start( ) > answer_plus.isize( ) - K ) break;
                    min_over = Min( min_over, last - cov[j].Start( ) );
                    if (cov_verbose)
                         out << cov[j] << " [" << last - cov[j].Start( ) << "]\n";
                    last = Max( last, cov[j].Stop( ) );    }
               out << "minimum overlap = " << min_over << "\n";    }

          // Return answer.

          return answer;    }
     else
     {    out << "\nFAILED!" << endl;
          return basevector( );    }    }

void AddToAssistedAssembly( basevector& assembly,
                            const basevector& assembly_addition,
                            const basevector& ref,
                            const unsigned ref_pos,
                            const unsigned ref_size,
                            vector<PositionLink>& assembly_reference_link ) {
  basevector ref_section;
  ref_section.SetToSubOf( ref, ref_pos, ref_size );

  if( assembly_addition == ref_section )
    assembly_reference_link.push_back( PositionLink( assembly.size(), ref_pos, ref_size ) );

  assembly = Cat( assembly, assembly_addition );
}


int main( int argc, char *argv[] )
{
     RunTime( );

     BeginCommandArguments;
     CommandArgument_String(HEAD);
     CommandArgument_String_OrDefault(PIPELINE, "");
     CommandArgument_String(REF);
     CommandArgument_Int_OrDefault(TRUNC, 0);
     CommandArgument_String_OrDefault(DIR, "");
     CommandArgument_String_OrDefault(POST, "");
     CommandArgument_String_OrDefault(RUN, "");
     // args above supplied by VAAL; args below supplied only when running this
     // module in stand-alone experimental mode
     CommandArgument_String_OrDefault_Doc(TRUE_REF, "",
          "if specified, used to assess assembly");
     CommandArgument_String_OrDefault(KXS, "{29,20,29}");
     CommandArgument_String_OrDefault(START, "");
     CommandArgument_Bool_OrDefault(PRINT_IDS, False);
     CommandArgument_Int_OrDefault_Doc(DUMP_NHOOD_READS, 0,
          "if 1, print nhood reads, including bases; if 2, include intensities");
     CommandArgument_String_OrDefault(OUT, "");
     CommandArgument_String_OrDefault(ASSEMBLY_SUFFIX, "");
     CommandArgument_Bool_OrDefault(ASSESS, False);
     CommandArgument_Int_OrDefault(EXTRA, 1000);
     EndCommandArguments;
     if ( PIPELINE.empty() ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

     // Sanity checks.

     if (ASSESS) ForceAssert( !TRUE_REF.empty() );

     // Define directories, etc.

     String workdir, workhead;
     VAALSetup( RUN, HEAD, REF, DIR, POST, workdir, workhead, TRUNC,
          PIPELINE );
     if ( OUT.empty() && START.empty() )
     {    if ( ASSEMBLY_SUFFIX.empty() ) OUT = workhead + ".assembly0.fastb";
          else OUT = workhead + "." + ASSEMBLY_SUFFIX + ".assembly0.fastb";    }
     ofstream xout;
     if ( START.empty() )
     {    cout << "writing to " << workhead + ".VAAL2.out\n" << endl;
          OpenOfstream( xout, "out", workhead + ".VAAL2.out" );    }
     ostream& out = ( START.empty() ? xout : cout );
     if ( START.empty() )
          command.PrintTheCommandPretty( out );

     const String LINK( workhead + ".assembly_ref.link" );

     // Load data.

     vecbasevector ref;
     String REFX;
     vec<String> HEADS;
     VAALLoadRef( HEAD, TRUNC, PIPELINE, workhead, REF, REFX, ref, HEADS );

     vecbvec bases;
     bases.ReadAll( workhead + ".trimmed.fastb" );

     String TRUE_REFX;
     if ( !TRUE_REF.empty() )
     {
         if ( !TRUE_REF.Contains( ".fasta", -1 ) )
              TRUE_REF += ".fasta";
         TRUE_REFX = VAALRefName(TRUE_REF,HEADS,workhead);
     }

     // Create searchable version of marks.

     VirtualMasterVec<PlacementMarkVec> marks( (workhead+".marks").c_str() );
     vec<placement_mark_plus> marks_plus;
     marks_plus.reserve( marks.sizeSum() );

     // Don't use reads that have multiple placements, far apart.
     // (NO, TURNED OFF.)
     // const int max_sep = 5000;
     for ( size_t id = 0; id < marks.size(); id++ )
     {
        PlacementMarkVec const& markVec = marks[id];
        /* Check for far-apart placements TURNED OFF
        Bool bad = False;
        for ( unsigned int j1 = 0; j1 < markVec.size(); j1++ )
        {
            for ( unsigned int j2 = j1 + 1; j2 < markVec.size(); j2++ )
            {
                const placement_mark &m1 = markVec[j1], &m2 = markVec[j2];
                if ( m1.Tig() != m2.Tig() || Abs((int) m1.Pos()
                        - (int) m2.Pos()) > max_sep )
                {
                    bad = True;
                }
            }
        }
        if ( bad )
            continue; */
        for ( unsigned int j = 0; j < markVec.size(); j++ )
            marks_plus.push_back(placement_mark_plus(id, markVec[j]));
     }
     Sort(marks_plus);

     // Load safe intervals.

     vecbitvector safe( workhead + ".safespots" );

     // Define values for KX.
     vec<int> kxs;
     ParseIntSet( KXS, kxs, False );

     // Define outputs
     vecbasevector assembly;
     ContigPositionLink assembly_reference_link( ref.size() ); // positions where assembled contig matches reference

     // Case 1: assemble single window.

     if ( !START.empty() && !START.Contains( "-" ) )
     {
          // Find interval to be assembled.

          int TIG = START.Before( "." ).Int( ), stop = START.After( "." ).Int( );
          while( safe[TIG][stop] )
          {    ++stop;
               ForceAssert( static_cast<unsigned>(stop) < safe[TIG].size( ) );    }
          while( !safe[TIG][stop] )
          {    ++stop;
               ForceAssert( static_cast<unsigned>(stop) < safe[TIG].size( ) );    }
          int startx = stop - 1;
          while(1)
          {    if ( safe[TIG][startx] ) break;
               ForceAssert( startx > 0 );
               --startx;    }
          startx -= 20;
          int Start = startx, STOP = stop + 20;

          vecbasevector extra_reads;
          assembly.resize(1);
          assembly_reference_link.resize(1);
          for ( int r = 0; r < kxs.isize( ); r++ )
          {    int min_over;
               assembly[0] = AssembleWindow( TIG, Start, STOP, ref, TRUE_REFX, REFX,
                    marks_plus, bases, PRINT_IDS, DUMP_NHOOD_READS, out, kxs[r],
                    min_over, ASSESS, EXTRA, extra_reads, workhead );
               assembly_reference_link[0].push_back( PositionLink( 0, Start, assembly[0].size() ) );
               if ( !assembly[0].empty( ) && min_over >= 20 ) break;
          }


     }

     // Case 2: assemble range.

     else if ( START.Contains( "-" ) )
     {    int TIG = START.Before( "." ).Int( );
          int start = START.Between( ".", "-" ).Int( );
          int stop = START.After( "-" ).Int( );
          while( !safe[TIG][start] )
          {    ForceAssert( start >= 0 );
               --start;    }
          while( !safe[TIG][stop] )
          {    ForceAssert( static_cast<unsigned>(stop) < safe[TIG].size( ) );
               ++stop;    }
          while( safe[TIG][start] )
          {    ForceAssert( start >= 0 );
               --start;    }
          ++start;
          while( safe[TIG][stop] )
          {    ForceAssert( static_cast<unsigned>(stop) < safe[TIG].size( ) );
               ++stop;    }
          out << "To assemble " << TIG << "." << start << "-" << stop << "\n\n";
          assembly.resize(1);
          assembly_reference_link.resize(1);
          for ( int pass = 1; pass <= 2; pass++ )
          {    if ( pass == 1 ) out << "preview:\n\n";
               else out << "\nassembly:\n\n";
               for ( int i = start; i < stop; i++ )
               {    int j, k;
                    for ( j = i + 1; j < stop; j++ )
                         if ( !safe[TIG][j] ) break;
                    out << "to copy " << TIG << "." << i << "-" << j << "\n";
                    basevector x;
                    if ( pass == 2 )
                    {    x.SetToSubOf( ref[TIG], i, j-i );
                         if ( i > start )
                              assembly[0].resize( assembly[0].isize( ) - 20 );
                         AddToAssistedAssembly( assembly[0], x, ref[TIG], i, j-i, assembly_reference_link[TIG] );    }
                    if ( j == stop ) break;
                    for ( k = j + 1; k < stop; k++ )
                         if ( safe[TIG][k] ) break;
                    out << "to assemble " << TIG << "." << j-20 << "-" << k+20
                         << " [" << k-j+40 << "]\n";
                    if ( pass == 2 )
                    {    int min_over;
                         basevector loc;
                         vecbasevector extra_reads;
                         for ( int r = 0; r < kxs.isize( ); r++ )
                         {    int min_over;
                              loc = AssembleWindow( TIG, j-20, k+20, ref, TRUE_REFX,
                                   REFX, marks_plus, bases, PRINT_IDS,
                                   DUMP_NHOOD_READS, out, kxs[r], min_over,
                                   ASSESS, EXTRA, extra_reads, workhead );
                              if ( !loc.empty( ) && min_over >= 20 ) break;
                              else if ( r == kxs.isize( ) - 1 )
                              {    loc.SetToSubOf(
                                        assembly[TIG], j-20, k-j+40 );    }    }
                         assembly[0].resize( assembly[0].isize( ) - 20 );
                         AddToAssistedAssembly( assembly[0], x, ref[TIG], j-20, k-j+40, assembly_reference_link[TIG] );    }
                    i = k - 1;    }    }
          out << "\nASSESSMENT OF ASSEMBLY:\n\n";
          // assembly[0].Print( out, "assembly" );
          {    Ofstream( aout, "assembly.fasta" );
               assembly[0].Print( aout, "assembly" );    }
          for ( int pass = 1; pass <= 2; pass++ )
          {    String GENOME;
               if ( pass == 1 ) GENOME = TRUE_REFX.Before( ".fasta" );
               else GENOME = REFX.Before( ".fasta" );
               if ( pass == 1 ) out << "\n";
               out << "ALIGNMENT TO ";
               if ( pass == 1 ) out << "SAMPLE GENOME:" << endl;
               else out << "RELATED REFERENCE:" << endl;
               SystemSucceed( "QueryLookupTable K=12 MM=12 MC=0.15 SEQS="
                    "assembly.fasta L=" + GENOME + ".lookup"
                    + " VISUAL=True NH=True QUIET=True" );    }    }

     // Case 3: assemble entire genome.

     else
     {    ForceAssert( START.empty() );

          assembly.resize( ref.size( ) );
          for ( size_t tig = 0; tig < ref.size( ); tig++ )
          {    for ( int i = 0; i < ref[tig].isize( ); i++ )
               {    int j = i + 1, k;
                    if ( i == 0 )
                    {    for ( j = i + 1; j < ref[tig].isize( ); j++ )
                              if ( safe[tig][j] ) break;    }
                    for ( ; j < ref[tig].isize( ); j++ )
                         if ( !safe[tig][j] ) break;
                    out << "to copy " << tig << "." << i << "-" << j << "\n";
                    basevector x;
                    x.SetToSubOf( ref[tig], i, j-i );
                    if ( i > 0 ) assembly[tig].resize( assembly[tig].isize( ) - 20 );

                    AddToAssistedAssembly( assembly[tig], x, ref[tig], i, j-i, assembly_reference_link[tig] );

                    if ( j == ref[tig].isize( ) ) break;
                    for ( k = j + 1; k < ref[tig].isize( ); k++ )
                         if ( safe[tig][k] ) break;

                    basevector loc;
                    unsigned ref_start, ref_size;

                    if ( k+20 > ref[tig].isize( ) )
                    {    out << "to copy " << tig << "." << j-20 << "-" << k
                              << "\n";
                      ref_start = j-20;
                      ref_size = k-j+20;
                      loc.SetToSubOf( ref[tig], ref_start, ref_size );

                    }
                    else
                    {    vecbasevector extra_reads;
                         ref_start = j-20;
                         ref_size = k-j+40;

                         out << "to assemble " << tig << "." << ref_start << "-" << ref_start+ref_size
                              << " [" << ref_size << "]" << endl;
                         for ( int r = 0; r < kxs.isize( ); r++ )
                         {    int min_over;
                              loc = AssembleWindow( tig, ref_start, ref_start+ref_size, ref, TRUE_REFX,
                                   REFX, marks_plus, bases, PRINT_IDS,
                                   DUMP_NHOOD_READS, out, kxs[r], min_over, ASSESS,
                                   EXTRA, extra_reads, workhead );

                              if ( !loc.empty( ) && min_over >= 20 ) break;
                              else if ( r == kxs.isize( ) - 1 )
                              {    loc.SetToSubOf(
                                        ref[tig], ref_start, ref_size );    }    }    }
                    assembly[tig].resize( assembly[tig].isize( ) - 20 );

                    AddToAssistedAssembly( assembly[tig], loc, ref[tig], ref_start, ref_size, assembly_reference_link[tig] );

                    i = k - 1;
               }
          }
     }

     // Write assembly.

     if ( !OUT.empty() ) assembly.WriteAll(OUT);
     BinaryWriter::writeFile( LINK.c_str(), assembly_reference_link );
}
