/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

/// VAAL3.  Read in assembly from VAAL2, mark trusted bases.  Also
/// change some bases in the assembly.
///
/// See VAAL1.cc for general info.

#include "Basevector.h"
#include "Bitvector.h"
#include "FeudalMimic.h"
#include "MainTools.h"
#include "Qualvector.h"
#include "feudal/BinaryStream.h"
#include "lookup/ImperfectLookup.h"
#include "lookup/LookAlign.h"
#include "math/Functions.h"
#include "math/HoInterval.h"
#include "paths/KmerBaseBroker.h"
#include "paths/KmerPath.h"
#include "paths/ReadsToPathsCoreX.h"
#include "paths/Unipath.h"
#include "polymorphism/VAALTools.h"
#include "solexa/SolexaTools.h"

int main( int argc, char *argv[] )
{
     RunTime( );

     BeginCommandArguments;
     CommandArgument_String(HEAD);
     CommandArgument_String(REF);
     CommandArgument_String_OrDefault( RUN, "" );
     CommandArgument_Int_OrDefault(TRUNC, 0);
     CommandArgument_String_OrDefault(DIR, "");
     CommandArgument_String_OrDefault(POST, "");
     CommandArgument_Int_OrDefault_Doc(QMIN, 10,
          "Define trusted read bases to be those having quality score >= QMIN." );
     CommandArgument_Bool_OrDefault_Doc(VERBOSE, True, "show base-by-base marking" );
     CommandArgument_String_OrDefault(ASSEMBLY_SUFFIX, "");
     CommandArgument_String_OrDefault(PIPELINE, "");
     EndCommandArguments;
     if ( PIPELINE == "" ) PIPELINE = Getenv("SOLEXA_PIPELINE_DIR");

    // Define directories, etc.

     double clock = WallClockTime( );
     String workdir, workhead;
     VAALSetup( RUN, HEAD, REF, DIR, POST, workdir, workhead, TRUNC,
          PIPELINE );
     cout << "writing to " << workhead + ".VAAL3.out\n" << endl;
     Ofstream( out, workhead + ".VAAL3.out" );
     command.PrintTheCommandPretty( out );

     // Load data.

     vecbasevector bases, ref;
     String REFX;
     VAALLoad( workhead, HEAD, ".vectrim", TRUNC, REF, REFX, bases, ref,
          PIPELINE );

     // bases.clear( );
     // bases.ReadAll( workhead + ".trimmed.fastb" );
     vecqualvector Q;
     if ( QMIN >= 0 )
     {    vec<String> HEADS;
          ExpandHead( HEAD, HEADS, TRUNC, PIPELINE );
          for ( int i = 0; i < HEADS.isize( ); i++ )
               Q.ReadAll( HEADS[i] + ".qualb", i > 0 );
          ForceAssertEq( Q.size( ), bases.size( ) );
          for ( vecqvec::size_type i = 0; i < Q.size( ); i++ )
               Q[i].resize( bases[i].size( ) );    }
     String workhead_plus = workhead;
     if ( ASSEMBLY_SUFFIX != "" ) workhead_plus += "." + ASSEMBLY_SUFFIX;
     String ASSEMBLY0 = workhead_plus + ".assembly0.fastb";
     String ASSEMBLY = workhead_plus + ".assembly.fastb";
     String TRUSTED = workhead_plus + ".assembly.trusted";
     vecbasevector assembly(ASSEMBLY0);
     int nreads = bases.size( ), ntigs = assembly.size( );

     const String LINK( workhead + ".assembly_ref.link" );
     ContigPositionLink assembly_reference_link( ref.size() );
     if( IsRegularFile( LINK ) )
         BinaryReader::readFile( LINK.c_str(), &assembly_reference_link );
     const String LINK_TRUSTED( LINK + ".trusted" );

     // Anchor reads on assembly.

     vecbasevector bases_assembly(bases);
     bases_assembly.Append(assembly);
     vecKmerPath paths, pathsrc;
     vec<big_tagged_rpint> pathsdb;
     int K = 28;
     cout << Date( ) << ": generating paths" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXXX
     ReadsToPathsCoreY( bases_assembly, K, paths, pathsrc, pathsdb );
     cout << Date( ) << ": placing reads" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
     vec<placement> places;
     AnchorReads( K, nreads, assembly, paths, pathsrc, pathsdb, places );
     vec<int> placements( nreads, 0 );
     for ( int i = 0; i < places.isize( ); i++ )
          ++placements[ places[i].id ];

     // If not placed with K = 28, try K = 20.  Finally, try alignment.

     {    int K = 20;
          vecbasevector bases2(bases);
          for ( int i = 0; i < nreads; i++ )
               if ( placements[i] > 0 ) bases2[i].resize(0);
          vecbasevector bases2_assembly(bases2);
          bases2_assembly.Append(assembly);
          ReadsToPathsCoreY( bases2_assembly, K, paths, pathsrc,
               pathsdb );
          vec<placement> places2;
          AnchorReads( K, nreads, assembly, paths, pathsrc, pathsdb, places2 );
          places.append(places2);
          for ( int i = 0; i < places2.isize( ); i++ )
          {    ++placements[ places2[i].id ];
               bases2[ places2[i].id ].resize(0);    }
          SystemSucceed( "MakeLookupTable SOURCE=\"" + ASSEMBLY0
               + "\" LOOKUP_ONLY=True " + "OUT_HEAD=\""
               + ASSEMBLY0.Before( ".fastb" ) + "\" SIMPLE_SOURCE=True" );
          vec<look_align> aligns;
          vec<int> min_errors;
          cout << Date( ) << ": starting alignment" << endl;
          ImperfectLookup( 12, bases2, workhead_plus + ".assembly0.lookup",
               aligns, min_errors, FW_OR_RC, 2, 4 );
          cout << Date( ) << ": done" << endl;
          for ( int i = 0; i < aligns.isize( ); i++ )
          {    const look_align& la = aligns[i];
               places.push( la.target_id, la.pos2( ), la.query_id, la.Fw1( ) );
               ++placements[la.query_id];    }    }

     // Remove reads placed multiply.

     vec<Bool> to_remove( places.size( ), False );
     for ( int i = 0; i < places.isize( ); i++ )
          if ( placements[ places[i].id ] > 1 ) to_remove[i] = True;
     EraseIf( places, to_remove );
     BinaryWriter::writeFile( workhead_plus + ".assembly.placements", places );

     // For every read anchored by a unique K-mer to the reference, that is not
     // anchored to the assisted assembly, find the "nearest" unique K-mer in the
     // reference that lies in the assisted assembly.  If that K-mer is also unique
     // in the assisted assembly, indirectly map the read to the assisted assembly.

     vec<placement> placesr;
     BinaryReader::readFile( workhead + ".reference.placements", &placesr );
     vecbasevector assembly_ref(assembly);
     assembly_ref.Append(ref);

     /*
     vecKmerPath xpaths, xpathsrc;
     vec<big_tagged_rpint> xpathsdb;
     ReadsToPathsCoreY( assembly_ref, K, xpaths, xpathsrc, xpathsdb );
     vec<placement> extra_places;
     for ( int i = 0; i < placesr.isize( ); i++ )
     {    const placement& pr = placesr[i];
          int id = pr.id;
          if ( placed[id] ) continue;
          const KmerPath& r = xpaths[ pr.tig + assembly.size( ) ];
          int posb = pr.pos + bases[id].isize( ) / 2;
          int posf = posb + 1;
          if ( posb < 0 || posf > ref[pr.tig].isize( ) - K ) continue;
          KmerPathLoc lb = r.Begin( ) + posb, lf = r.Begin( ) + posf;
          Bool bdone = False, fdone = False;
          Bool succeed = False;
          while(1)
          {    static vec<longlong> con;
               for ( int pass = 1; pass <= 2; pass++ )
               {    if ( pass == 1 && bdone ) continue;
                    if ( pass == 2 && fdone ) continue;
                    longlong X = ( pass == 1 ? lb : lf ).GetKmer( );
                    Contains( xpathsdb, X, con );
                    int count = 0, apos = 0, rpos = 0, aid = 0;
                    Bool fw = True;
                    for ( int j = 0; j < con.isize( ); j++ )
                    {    const big_tagged_rpint& t = xpathsdb[ con[j] ];
                         aid = t.ReadId( );
                         if ( aid >= assembly.size( ) ) continue;
                         if ( ++count >= 2 ) break;
                         const KmerPath& A
                              = ( t.Fw( ) ? xpaths[aid] : xpathsrc[aid] );

                         // apos is the position of the kmer X in the assisted
                         // assembly or its reverse complement

                         apos = X - t.Start( );
                         for ( int u = 0; u < t.PathPos( ); u++ )
                              apos += A.Length(u);

                         // rpos is the position of X on the reference

                         rpos = ( pass == 1 ? posb : posf );
                         fw = t.Fw( );
                         break;    }
                    if ( count == 1 )
                    {    int read_start_on_assembly = apos + pr.pos - rpos;
                         if ( !fw )
                         {    read_start_on_assembly =
                                   assembly[aid].isize( )
                                   - ( read_start_on_assembly
                                        + bases[id].isize( ) );    }
                         extra_places.push( aid, read_start_on_assembly, id,
                              !( !fw ^ !pr.fw ) );
                         ++placements[id];
                         succeed = True;
                         goto end_of_placement_loop;    }
                    if ( count >= 2 ) break;    }
               if ( lb.atBegin( ) ) bdone = True;
               else
               {    lb = lb - 1;
                    --posb;    }
               if ( lf.atEnd( ) ) fdone = True;
               else
               {    lf = lf + 1;
                    ++posf;    }
               if ( bdone && fdone ) break;    }
          end_of_placement_loop: continue;    }
     places.append(extra_places);
     */

     // For every 5-mer in every placed read, we require that there is another
     // read 5-mer placed in the same position that agrees with it at every trusted
     // base on the first 5-mer.  Otherwise, we delete the first placement.

     const int window = 5;
     for ( int pass = 1; pass <= 2; pass++ )
     {
     vec<Bool> reject( places.size( ), False );
     vec< vec< vec<int> > > allcallsMi(ntigs), allcallsMu(ntigs);
     for ( int i = 0; i < ntigs; i++ )
     {    allcallsMi[i].resize( assembly[i].size( ) );
          allcallsMu[i].resize( assembly[i].size( ) );    }
     for ( int i = 0; i < places.isize( ); i++ )
     {    int id = places[i].id, tig = places[i].tig, start = places[i].pos;
          Bool fw = places[i].fw;
          int n = bases[id].size( );
          for ( int u = 0; u < n; u++ )
          {    int pos = ( fw ? u : n - u - 1 );
               char base = ( fw ? bases[id][pos] : 3 - bases[id][pos] );
               if ( start+u >= 0 && start+u <= assembly[tig].isize( ) - window )
               {    allcallsMi[tig][start+u].push_back(i);
                    allcallsMu[tig][start+u].push_back(u);    }    }    }
     for ( int i = 0; i < places.isize( ); i++ )
     {    int id = places[i].id, tig = places[i].tig, start = places[i].pos;
          Bool fw = places[i].fw;
          int n = bases[id].size( );
          for ( int u = 0; u < n; u++ )
          {    int pos = ( fw ? u : n - u - 1 );
               if ( fw && pos > n - window ) continue;
               if ( !fw && pos < window - 1 ) continue;
               if ( start+u >= 0 && start+u <= assembly[tig].isize( ) - window )
               {    Bool confirmed = False;
                    for ( int x = 0; x < allcallsMi[tig][start+u].isize( ); x++ )
                    {    int i2 = allcallsMi[tig][start+u][x];
                         if ( i2 == i ) continue;
                         int u2 = allcallsMu[tig][start+u][x];
                         int id2 = places[i2].id, start2 = places[i2].pos;
                         Bool fw2 = places[i2].fw;
                         int n2 = bases[id2].size( );
                         int pos2 = ( fw2 ? u2 : n2 - u2 - 1 );
                         if ( fw2 && pos2 + (window-1) >= n2 ) continue;
                         if ( !fw2 && pos2 - (window-1) < 0 ) continue;
                         Bool agree = True;
                         for ( int y = 0; y < window; y++ )
                         {    int posy = pos + ( fw ? y : -y );
                              int pos2y = pos2 + ( fw2 ? y : -y );
                              char basey = ( fw
                                   ? bases[id][posy] : 3 - bases[id][posy] );
                              char base2y = ( fw2
                                   ? bases[id2][pos2y] : 3 - bases[id2][pos2y] );
                              Bool good = False;
                              if ( Q[id][posy] >= QMIN ) good = True;
                              if ( !good ) continue;
                              if ( basey != base2y ) { agree = False; break; }    }
                         if (agree) { confirmed = True; break; }    }
                    if ( !confirmed ) { reject[i] = True; break; }    }    }    }
     EraseIf( places, reject );
     }
     BinaryWriter::writeFile( workhead_plus + ".assembly.placements_plus", places );

     //
     // Write a fasta file of the reads that were filtered out because:
     // 1) They were never initially placed.
     // 2) They were placed multiply.
     // 3) They were not part of an overlapping 5-mer.
     //
     vec<Bool> placed( nreads, False );
     for ( int i = 0; i < places.isize( ); i++ )
          placed[ places[i].id ] = True;

     Ofstream( unplaced, workhead + ".unplaced.fasta" );
     for( unsigned read_id = 0; read_id < placed.size( ); read_id++ ) {
       if( !placed[read_id] ) {
         ostringstream s;
         s << read_id;
         bases[read_id].Print( unplaced, s.str() );
       }
     }
     unplaced.close();

     // Paint the assembly.

     cout << Date( ) << ": painting assembly" << endl; // XXXXXXXXXXXXXXXXXXXXXXXXXX
     vec< vec< vec<char> > > calls(ntigs), allcalls(ntigs);
     vec< vec< vec<char> > > callsM(ntigs), allcallsM(ntigs);
     for ( int i = 0; i < ntigs; i++ )
     {    calls[i].resize( assembly[i].size( ) );
          allcalls[i].resize( assembly[i].size( ) );
          callsM[i].resize( assembly[i].size( ) );
          allcallsM[i].resize( assembly[i].size( ) );    }
     for ( int i = 0; i < places.isize( ); i++ )
     {    int id = places[i].id, tig = places[i].tig, start = places[i].pos;
          Bool fw = places[i].fw;
          int n = bases[id].size( );
          for ( int u = 0; u < n; u++ )
          {    int pos = ( fw ? u : n - u - 1 );
               char base = ( fw ? bases[id][pos] : 3 - bases[id][pos] );
               if ( start+u >= 0 && start+u < calls[tig].isize( ) )
               {    allcallsM[tig][start+u].push_back(base);
                    if ( placements[id] == 1 )
                         allcalls[tig][start+u].push_back(base);
                    Bool good = False;
                    if ( Q[id][pos] >= QMIN ) good = True;
                    if (good)
                    {    callsM[tig][start+u].push_back(base);
                         if ( placements[id] == 1 )
                         {    calls[tig][start+u].push_back(
                                   base );    }    }    }    }    }

     // Edit assembly and mark bases as trusted.  Find all the read bases lying over
     // a given position on the assembly; some are trusted and some are not.  Let
     // n1 be the highest count amongst trusted bases, and let n2 be the second
     // highest count.  Similarly, let N1 be the highest count amongst all read
     // bases and let N2 be the second highest count.  We require that
     // - n1 >= 5;
     // - n1 >= 5 * n2;
     // - N1 >= 2 * N2;
     // - n1 and N1 are counting the same base b.
     // In this case, we set the assembly base to b (which it usually is already),
     // and mark it as trusted.
     // Added requirement.  It has to work whether multiply-placed reads are
     // included or not.

     cout << Date( ) << ": marking trusted bases" << endl; // XXXXXXXXXXXXXXXXXXXXXX
     vecbitvector covered, trusted, dangerous;
     Mimic(assembly, covered), Mimic(assembly, trusted), Mimic(assembly, dangerous);
     vecbitvector trustedw;
     Mimic(assembly,trustedw);
     for ( int i = 0; i < ntigs; i++ )
     {    if (VERBOSE)
               out << "\ncontig " << i << ", len = " << assembly[i].size( ) << "\n";

       vector<PositionLink>::iterator link_iter = assembly_reference_link[i].begin();
       vector<PositionLink>::iterator link_end = assembly_reference_link[i].end();

          double mean_cov = 0.0;
          for ( int j = 0; j < assembly[i].isize( ); j++ )
               mean_cov += allcalls[i][j].size( );
          mean_cov /= double( assembly[i].size( ) );
          for ( int j = 0; j < assembly[i].isize( ); j++ )
          {    vec<int> count( 4, 0 ), allcount( 4, 0 );
               vec<int> countM( 4, 0 ), allcountM( 4, 0 );
               for ( int u = 0; u < calls[i][j].isize( ); u++ )
                    ++count[ calls[i][j][u] ];
               for ( int u = 0; u < allcalls[i][j].isize( ); u++ )
                    ++allcount[ allcalls[i][j][u] ];
               for ( int u = 0; u < callsM[i][j].isize( ); u++ )
                    ++countM[ callsM[i][j][u] ];
               for ( int u = 0; u < allcallsM[i][j].isize( ); u++ )
                    ++allcountM[ allcallsM[i][j][u] ];
               Bool T = False, TM = False;

               int base;
               for ( base = 0; base < 4; base++ )
               {    if ( count[base] >= 5 )
                    {    Bool bad = False;
                         for ( int x = 0; x < 4; x++ )
                         {    if ( x == base ) continue;
                              if ( !( count[base] >= 5*count[x] ) ) bad = True;    }
                         if ( !bad )
                         {    T = True;
                              break;    }    }    }
               if ( base == 4 ) base = assembly[i][j];
               for ( int x = 0; x < 4; x++ )
               {    if ( x == base ) continue;
                    if ( !( allcount[base] >= 2 * allcount[x] ) ) T = False;    }

               for ( int x = 0; x < 4; x++ )
               {    if ( x == base ) continue;
                    if ( !( count[base] >= 5 * count[x] ) )
                         dangerous[i].Set( j, True );    }
               for ( int x = 0; x < 4; x++ )
               {    if ( x == base ) continue;
                    if ( !( countM[base] >= 5 * countM[x] ) )
                         dangerous[i].Set( j, True );    }

               int baseM;
               for ( baseM = 0; baseM < 4; baseM++ )
               {    if ( countM[baseM] >= 5 )
                    {    Bool bad = False;
                         for ( int x = 0; x < 4; x++ )
                         {    if ( x == baseM ) continue;
                              if ( !( countM[baseM] >= 5*countM[x] ) )
                                   bad = True;    }
                         if ( !bad )
                         {    TM = True;
                              break;    }    }    }
               if ( baseM == 4 ) baseM = assembly[i][j];
               for ( int x = 0; x < 4; x++ )
               {    if ( x == baseM ) continue;
                    if ( !( allcountM[baseM] >= 2 * allcountM[x] ) ) TM = False;    }

               if ( T && TM && base == baseM
                    && allcalls[i][j].size( ) <= 3.0 * mean_cov )
               {    trusted[i].Set( j, True );
                    trustedw[i].Set( j, True );
                    if (VERBOSE)
                    {    if ( assembly[i][j] != base )
                         {    out << "correcting assembly[" << i << "][" << j
                                   << "]: " << as_base( assembly[i][j] ) << " --> "
                                   << as_base(base) << "\n";    }    }
                    assembly[i].Set( j, base );    }

               covered[i].Set( j, allcount[base] > 0 );

               if (VERBOSE)
               {    out << i << "." << j << " ";
                    if ( j >= 20 )
                    {    for ( int k = 0; k < 20; k++ )
                              out << as_base( assembly[i][j-20+k] );
                         out << " ";    }
                    out << as_base(base) << "[" << count[base] << "]";
                    for ( int x = 0; x < 4; x++ )
                    {    if ( x == base || count[x] == 0 ) continue;
                         out << " " << as_base(x) << "[" << count[x] << "]";    }
                    while( link_iter != link_end &&
                           unsigned(j) > link_iter->SourceStop() ) ++link_iter;
                    if( link_iter != link_end &&
                        unsigned(j) >= link_iter->SourcePos &&
                        unsigned(j) <= link_iter->SourceStop() ) {
                      unsigned int ref_pos = j - link_iter->SourcePos + link_iter->TargetPos;
                      out << " ref=" << i << "." << ref_pos << " [" << as_base(ref[i][ref_pos]) << "]";
                    }
                    else
                      out << " unmatched";

                    if ( !T ) out << " untrusted";
                    if ( dangerous[i][j] ) out << " (dangerous)";
                    out << "\n";    }    }

          int dang_radius = 20;
          for ( unsigned int j = dang_radius; j < trusted[i].size( ) - dang_radius; j++ )
          {    if ( dangerous[i][j] )
               {    for ( int k = -dang_radius; k <= dang_radius; k++ )
                         trusted[i].Set( j + k, False );    }    }    }

     trusted.WriteAll( TRUSTED + "z" );
     trustedw.WriteAll( TRUSTED + "w" );

     // Every 12-mer in the assembly must be contained in a read lying over it.
     // Otherwise we mark the 12-mer and 10 bases to the left and right as untrusted.

     const int Ksupp = 12, Kflank = 10;
     vecbitvector supported;
     Mimic( assembly, supported );
     for ( int i = 0; i < places.isize( ); i++ )
     {    int id = places[i].id, tig = places[i].tig, start = places[i].pos;
          Bool fw = places[i].fw;
          int n = bases[id].size( );
          int low = Max(0, -start), high = Min( n, assembly[tig].isize( ) - start );
          int matchcount = 0;
          for ( int u = low; u < high; u++ )
          {    int pos = ( fw ? u : n - u - 1 );
               char base = ( fw ? bases[id][pos] : 3 - bases[id][pos] );
               if ( assembly[tig][start+u] != base ) matchcount = 0;
               else
               {    ++matchcount;
                    if ( matchcount >= Ksupp )
                         supported[tig].Set( start+u-Ksupp+1, True );    }    }    }
     for ( size_t tig = 0; tig < assembly.size( ); tig++ )
     {    int n = assembly[tig].size( ), left, right;
          for ( left = 0; left < n; left++ )
               if ( covered[tig][left] ) break;
          for ( right = 0; right < n; right++ )
               if ( covered[tig][ n - right - 1 ] ) break;
          for ( int j = left; j <= n - Ksupp - right; j++ )
          {    if ( !supported[tig][j] )
               {    if (VERBOSE)
                    {    out << "bracketing as untrusted " << tig << "."
                              << Max( 0, j-Kflank ) << "-"
                              << Min( n, j+Ksupp+Kflank ) << "\n";    }
                    for ( int u = -Kflank; u < Ksupp+Kflank; u++ )
                    {    if ( j+u >= 0 && j+u < n )
                              trusted[tig].Set( j+u, False );    }    }    }    }

     // Find repeat joins:
     // 1. For each placed read, effectively trim off bases on the end until the
     // read aligns perfectly.  This info is captured by "bracket".
     // ...
     // Do not trust bases in repeat joins.

     vec< vec<ho_interval> > bracket(ntigs);
     for ( int i = 0; i < places.isize( ); i++ )
     {    int id = places[i].id, tig = places[i].tig, start = places[i].pos;
          Bool fw = places[i].fw;
          int n = bases[id].size( ), u;
          int low = Max(0, -start), high = Min( n, assembly[tig].isize( ) - start );
          if (fw)
          {    for ( u = low; u < high; u++ )
               {    int pos = ( fw ? u : n - u - 1 );
                    char base = ( fw ? bases[id][pos] : 3 - bases[id][pos] );
                    if ( assembly[tig][start+u] != base ) break;    }
               bracket[tig].push( start+low, start+u );    }
          else
          {    for ( u = high-1; u >= low; u-- )
               {    int pos = ( fw ? u : n - u - 1 );
                    char base = ( fw ? bases[id][pos] : 3 - bases[id][pos] );
                    if ( assembly[tig][start+u] != base ) break;    }
               bracket[tig].push( start+u+1, start+high );    }    }
     const int min_width = 12, max_width = 50;
     vecKmerPath ypaths, ypathsrc;
     vec<big_tagged_rpint> ypathsdb;
     ReadsToPathsCoreY( assembly_ref, min_width,
          ypaths, ypathsrc, ypathsdb );
     for ( int t = 0; t < ntigs; t++ )
     {    int M = 0;
          vec<ho_interval>& B = bracket[t];
          for ( int j = 0; j < B.isize( ); j++ )
               M = Max( M, B[j].Length( ) );
          UniqueSort(B);
          int bstart = 0, bstop = 0;
          KmerPathLoc L = ypaths[t].Begin( );
          for ( int p = 0; p < assembly[t].isize( ); p++ )
          {    while( bstart < B.isize( ) && B[bstart].Start( ) + M <= p ) ++bstart;
               while( bstop < B.isize( ) && B[bstop].Start( ) < p + max_width )
                    ++bstop;
               for ( int w = min_width; w <= max_width; w++ )
               {    int left = p, right = p + w;
                    Bool across = False;
                    for ( int j = bstart; j < bstop; j++ )
                    {    if ( B[j].Start( ) < left && B[j].Stop( ) > right )
                         {    across = True; break;    }    }
                    if (across) continue;
                    longlong x = L.GetKmer( );
                    static vec<longlong> con;
                    Contains( ypathsdb, x, con );
                    int locs = 0;
                    for ( int j = 0; j < con.isize( ); j++ )
                    {    const big_tagged_rpint& T = ypathsdb[ con[j] ];
                         int id = T.ReadId( );
                         if ( id < ntigs ) continue;
                         const KmerPath& p = ( T.Fw( ) ? ypaths[id] : ypathsrc[id] );
                         int pp = T.PathPos( );
                         KmerPathLoc M1 = L, M2( p, pp, x - p.Start(pp) );
                         Bool agree = True;
                         for ( int u = 0; u <= w - min_width; u++ )
                         {    if ( M1.GetKmer( ) != M2.GetKmer( ) )
                              {    agree = False; break;    }
                              if ( u < w - min_width
                                   && ( M1.atEnd( ) || M2.atEnd( ) ) )
                              {    agree = False; break;    }
                              if ( u < w - min_width )
                              {    M1 = M1 + 1, M2 = M2 + 1;    }    }
                         if (agree) ++locs;    }
                    if ( locs <= 1 ) break;
                    if (VERBOSE)
                    {    out << "repeat join at " << t << "." << p << "-"
                              << p+w << "\n";    }
                    for ( int u = 0; u < w; u++ )
                         trusted[t].Set( p+u, False );    }
               if ( L.atEnd( ) ) break;
               L = L + 1;    }    }

     // Generate file indicating whether links between the assisted assembly and the ref are trusted.
     vecbitvector link_trusted;
     Mimic( ref, link_trusted, False );

     for( size_t tig = 0; tig < ref.size(); tig++ ) {
       sort( assembly_reference_link[tig].begin(), assembly_reference_link[tig].end(), PositionLink::cmp_target_pos );
       vector<PositionLink>::iterator link_iter = assembly_reference_link[tig].begin();
       vector<PositionLink>::iterator link_end = assembly_reference_link[tig].end();
       for( unsigned pos = 0; pos < ref[tig].size(); pos++ ) {
         while( link_iter != link_end && pos > link_iter->TargetStop() ) link_iter++;
         if( link_iter == link_end ) break;

         if( link_iter->IsInTarget( pos ) )
           link_trusted[tig].Set( pos, True );
       }
     }
     link_trusted.WriteAll( LINK_TRUSTED );

     // Write the assembly.

     trusted.WriteAll(TRUSTED);
     assembly.WriteAll(ASSEMBLY);
     Ofstream( tout, TRUSTED + ".txt" );
     Ofstream( aout, ASSEMBLY.RevBefore( "b" ) + "a" );
     for ( size_t i = 0; i < assembly.size( ); i++ )
     {    trusted[i].PrintFastaStyle( tout, "contig_" + ToString(i) );
          assembly[i].Print( aout, "contig_" + ToString(i) );    }
     out << ToString( 100.0 * Coverage(trusted), 1 ) << "% of bases are trusted"
          << "\n";
     out << "\n" << TimeSince(clock) << " used" << endl;
     SystemSucceed( "MakeLookupTable SOURCE=\"" + ASSEMBLY + "\" LOOKUP_ONLY=True "
          + "OUT_HEAD=\"" + ASSEMBLY.Before( ".fastb" )
          + "\" SIMPLE_SOURCE=True" );    }
