/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2007) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

/**
   \file

   This program takes the COMPOUND_FILE output of two TruePoly and/or 
   VAAL4 output and outputs the false positives, events not found, and in
   a separate file the number of true subevents per even and number found
   
*/

#include "MainTools.h"
#include "Basevector.h"
#include "String.h"
#include "TokenizeString.h"

#include <vector>
#include <map>
#include <algorithm>
#include <fstream>
#include <stdlib.h>
#include <math.h>


class SingleEvent {
public:
    int contig;
    int pos;
    String event;
    void clear( ) { contig = -1;  pos = -1;  event = ""; }
    friend ostream& operator<< ( ostream& out, const SingleEvent& s ) {
       return out << s.contig << " " << s.pos << " " << s.event << endl;
    }
};

class CompoundEvent {
public:
    SingleEvent orgEvent;
    vec<SingleEvent> events;
    void clear ( ) { orgEvent.clear( ); events.clear( ); }
};



void ReadEvents(String& theFile, vec<CompoundEvent>& theEvents) {
    Ifstream(fin, theFile);
    CompoundEvent cEvent;
    SingleEvent sEvent;
    bool first = true;
    while (true) {
       String line, afterColon;
       vec<String> tokens;
       getline(fin, line);
       if (!fin) break;
       if (line.Contains("Group")) {
           if (first) first = false;
           else {
              theEvents.push_back(cEvent);
              cEvent.clear();
           }
           afterColon = line.After(":");
           DeleteLeadingWhiteSpace(afterColon);
           Tokenize(afterColon, tokens);
           cEvent.orgEvent.contig = tokens[0].Int( );
           cEvent.orgEvent.pos = tokens[1].Int( );
           cEvent.orgEvent.event = tokens[2] + " " + tokens[3] + " " + tokens[4] + " " + tokens[5];
       } else if (line != "") {
           Tokenize(line, tokens);
           sEvent.clear( );
           sEvent.contig = tokens[0].Int ( );
           sEvent.pos = tokens[1].Int( );
           sEvent.event = tokens[2] + " " + tokens[3] + " " + tokens[4] + " " + tokens[5];
           cEvent.events.push_back(sEvent);
       }
    }
    // Push back final cEvent.  
    // Previous ones only pushed back after a line with "Group" in it.
    if (cEvent.events.size( ) > 0)
        theEvents.push_back(cEvent);
    fin.close();
}


void findEvent(SingleEvent& testEvent, vec<CompoundEvent>& checkEvents,
               int& group, int& eventNum, bool checkEvent = false) {
   // If none is found, group = -1
   group = -1;
   int testContig = testEvent.contig;
   int testPos = testEvent.pos;
   String doEvent;
   if (checkEvent) doEvent = testEvent.event;
   bool found = false;
   unsigned int ii = 0, jj = 0;
   for (ii = 0; ii < checkEvents.size( ) && !found; ++ii) {
      CompoundEvent& thisCompound = checkEvents[ii];
      vec<SingleEvent>& theEvents = thisCompound.events;
      for (jj = 0; jj < theEvents.size( ) && !found; ++jj) {
         SingleEvent& theSingleEvent = theEvents[jj];
         if (theSingleEvent.contig == testContig &&
             theSingleEvent.pos == testPos) found = true;
         if (checkEvent && (theSingleEvent.event != doEvent)) found = false;
      }
   }
   if (found) {
      group = ii-1;
      eventNum = jj-1;
   }
} 



bool isASNP(SingleEvent& theEvent) {
    String event = theEvent.event;
    String firstPartA = event.After("sample=");
    String firstPart = firstPartA.Before(" ref=");
    String secondPartA = event.After("ref=");
    String secondPart = secondPartA.Before(" right=");
 
    bool result = true;

    if ( firstPart != "A" && firstPart != "C" && 
         firstPart != "G" && firstPart != "T") 
       result = false;
    if ( secondPart != "A" && secondPart != "C" && 
         secondPart != "G" && secondPart != "T") 
       result = false;

    return result;
}



int main( int argc, char *argv[] ) 
{
  RunTime( );

  BeginCommandArguments;
  CommandArgument_String(COMPOUND_ASSEMBLY);
  CommandArgument_String(COMPOUND_TRUEPOLY);
  CommandArgument_String(EVENTS_FILE);
  CommandArgument_String(SUBEVENTS_FILE);
  EndCommandArguments;

  vec<CompoundEvent> theRefEvents, theReadsEvents;

  ReadEvents(COMPOUND_TRUEPOLY, theRefEvents);
  ReadEvents(COMPOUND_ASSEMBLY, theReadsEvents);


  // First do false positives into EVENTS_FILE
  Ofstream(efout, EVENTS_FILE);
  efout << "False positives\n";
  int numFalse = 0, group, eventNum;
  for (unsigned int ii = 0; ii < theReadsEvents.size( ); ++ii) {
     CompoundEvent& theCurrentCompound = theReadsEvents[ii];
     for (unsigned int jj = 0; jj < theCurrentCompound.events.size( ); ++jj) {
        SingleEvent& theCurrentEvent = theCurrentCompound.events[jj];
        findEvent(theCurrentEvent, theRefEvents, group, eventNum);
        if (group == -1) {
           efout << theCurrentEvent;
           numFalse++;
        }
     }
  }
  efout << "Number of false positives = " << numFalse << endl << endl;

  efout << endl << "--------------------------------" << endl << endl;


  // Below is for SUBEVENTS_FILE
  vec< pair<int, int> > trueSubeventsFound;


  // Next do sensitivity
  efout << "Sensitivity\n";
  for (unsigned int ii = 0; ii < theRefEvents.size( ); ++ii) {
     CompoundEvent& theCurrentCompound = theRefEvents[ii];
     efout << "Group " << ii << " (" << theCurrentCompound.events.size( ) << " subevents)" << endl;
     int numMissing = 0;
     vec<SingleEvent> missingEvents;
     for (unsigned int jj = 0; jj < theCurrentCompound.events.size( ); ++jj) {
        SingleEvent& theCurrentEvent = theCurrentCompound.events[jj];
        findEvent(theCurrentEvent, theReadsEvents, group, eventNum);
        if (group == -1) {
           numMissing++;
           missingEvents.push_back(theCurrentEvent);
        }
     }
     efout << "   " << numMissing << " subevents not found\n";
     for (unsigned int jj = 0; jj < missingEvents.size( ); ++jj) 
        efout << "      " << missingEvents[jj];
     efout << endl;
     pair<int, int> tempPair( theCurrentCompound.events.size( ),
                              theCurrentCompound.events.size( ) - numMissing );
     if ( theCurrentCompound.events.size( ) == 1 &&
          isASNP( theCurrentCompound.events[0] ) ) {
        tempPair.first = 1;   // Just put here to enforce the fact this is a SNP
     } else if ( theCurrentCompound.events.size( ) == 1 ) {
        tempPair.first = -1;  // Indiciates this is not a SNP
     }
     trueSubeventsFound.push_back(tempPair);
  }

  efout.close();

  // Finally do true [sub]events found in SUBEVENTS_FILE
  ReverseSort(trueSubeventsFound);
  Ofstream(sefout, SUBEVENTS_FILE); 
  sefout << "True Events\n";
  sefout << "# subevents\t\t# found\n";
  int detectable = 0, detected = 0, partial = 0;
  int SNPs = 0, SNPsFound = 0;
  int otherSingletons = 0, othersFound = 0;
  for (unsigned int ii = 0; ii < trueSubeventsFound.size( ); ++ii) {
     pair<int, int>& tempPair = trueSubeventsFound[ii];
     sefout << tempPair.first << "\t\t\t" << tempPair.second << endl;
     if (tempPair.first > 1) {
        detectable++;
        if (tempPair.second == tempPair.first) 
           detected++;
        else if (tempPair.second > 0)
           partial++;
     } else if (tempPair.first == 1) {
        SNPs++;
        if (tempPair.second == 1) SNPsFound++;
     } else if (tempPair.first == -1) {
        otherSingletons++;
        if (tempPair.second == 1) othersFound++;
     }
  }
  sefout << "#" << SNPs << " detectable SNPs\n";
  sefout << "#" << SNPsFound << " detected SNPs\n";
  sefout << "#" << otherSingletons << " detectable single event indels\n";
  sefout << "#" << othersFound << " detected single event indels\n";
  sefout << "#" << detectable << " detectable composites\n";
  sefout << "#" << detected << " detected composites\n";
  sefout << "#" << partial << " partially detected composites\n";
  sefout.close();

}
