/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

#include "Basevector.h"
#include "MainTools.h"
#include "Vec.h"
#include "paths/HyperKmerPath.h"
#include "paths/KmerPath.h"
#include "paths/ReadsToPathsCoreX.h"
#include "paths/Unipath.h"
#include "random/SequenceEntropy.h"
#include "polymorphism/VAALTools.h"
#include "solexa/SolexaTools.h"

// AnchorReads.  Place reads on reference, anchoring using kmers that are unique
// in the reference.
//
// INPUTS:
// paths = KmerPaths generated from concatenation of reads (size = nreads) 
//         and reference (size = nref)
// OUTPUT:
// places =  read placements.

void AnchorReads( const int K, const int nreads, const vecbasevector& ref,
     const vecKmerPath& paths, const vecKmerPath& pathsrc, 
     const vec<big_tagged_rpint>& pathsdb, vec<placement>& places,
     const double dinuke_entropy_floor )
{    int nref = ref.size( );
     vecKmerPath gpaths;
     vec<big_tagged_rpint> gpathsdb;
     for ( size_t i = nreads; i < paths.size( ); i++ )
          gpaths.push_back( paths[i] );
     CreateDatabase( gpaths, gpathsdb );
     vec< vec<int> > gpath_starts(nref);
     for ( int i = 0; i < nref; i++ )
     {    int pos = 0;
          for ( int j = 0; j < gpaths[i].NSegments( ); j++ )
          {    gpath_starts[i].push_back(pos);
               pos += gpaths[i].Length(j);    }    }
     places.clear( );
     for ( int tig = 0; tig < nref; tig++ )
     {    const KmerPath& g = paths[ tig + nreads ];
          int POS = 0;
          for ( int i = 0; i < g.NSegments( ); i++ )
          {    for ( longlong j = g.Start(i); j <= g.Stop(i); j++ )
               {    if ( dinuke_entropy_floor > 0.0 )
                    {    static basevector b;
                         b.SetToSubOf( ref[tig], POS++, K );
                         if ( DinukeEntropy(b) < dinuke_entropy_floor ) 
                              continue;    }
                    static vec<longlong> con;
                    Contains( pathsdb, j, con );
                    int ng = 0;
                    for ( int u = 0; u < con.isize( ); u++ )
                    {    const big_tagged_rpint& t = pathsdb[ con[u] ];
                         if ( t.ReadId( ) >= nreads ) ++ng;    }
                    if ( ng != 1 ) continue;
                    for ( int u = 0; u < con.isize( ); u++ )
                    {    const big_tagged_rpint& t = pathsdb[ con[u] ];
                         int id = t.ReadId( );
                         if ( id >= nreads ) continue;
                         const KmerPath& p = ( t.Fw( ) ? paths[id] : pathsrc[id] );
                         longlong gpos 
                              = gpath_starts[tig][i] - g.Start(i) + t.Start( );
                         for ( int v = 0; v < t.PathPos( ); v++ )
                              gpos -= p.Length(v);
                         if ( gpos < 0 ) continue;
                         places.push( tig, gpos, id, t.Fw( ) );    }    }    }    }
     UniqueSort(places);    }

void VAALSetup( String RUN, String& HEAD, String& REF, String& PRE, 
     String& POST, String& workdir, String& workhead, int& TRUNC,
          const String& PIPELINE )
{    if ( RUN == "" )
     {    char* username = getenv( "LOGNAME" );
          if ( username != 0 ) RUN = username;
          else FatalErr( "Unable to determine RUN." );    }
     vec<String> HEADS;
     ExpandHead( HEAD, HEADS, TRUNC, PIPELINE );
     if ( HEADS.solo( ) )
     {    if ( PRE == "" ) PRE = HEADS[0].RevBefore( "/" );
          if ( POST == "" ) POST = HEADS[0].RevAfter( "/" );    }
     else if ( HEAD.Freq( "." ) == 1 )
     {    String pretrunc = HEAD;
          if ( pretrunc.Contains( "[" ) ) pretrunc = pretrunc.Before( "[" );
          if ( PRE == "" ) PRE = HEADS[0].RevBefore( "/" );
          if ( POST == "" ) POST = pretrunc;    }
     else
     {    PRE = PIPELINE;
          PRE += "/mixed";
          if ( TRUNC != 0 ) PRE += "/trunc_" + ToString(TRUNC);
          Mkdir777(PRE);
          POST = HEAD;    }
     String REF0 = REF;
     if ( REF.Contains( ".edit1", -1 ) ) REF0 = REF.Before( ".edit1" );
     workdir = PRE + "/vs_" + REF0 + "/" + RUN;
     Mkdir777( PRE + "/vs_" + REF0 );
     Mkdir777(workdir);
     workhead = workdir + "/" + POST;
     REF += ".fasta";    }

void VAALLoad( const String& workhead, const String& HEAD, 
     const String& HEAD_TAIL, const int trunc, const String& REF, String& REFX, 
     vecbasevector& bases, vecbasevector& ref, const String& PIPELINE )
{    vec<String> HEADS;
     VAALLoadRef( HEAD, trunc, PIPELINE, workhead, REF, REFX, ref, HEADS );
     bases.clear();
     for ( size_t i = 0; i < HEADS.size( ); i++ )
          bases.ReadAll( HEADS[i] + HEAD_TAIL + ".fastb", true ); }

String VAALRefName( String const& REF,
                    vec<String> const& HEADS,
                    String const& workhead )
{
    String REFX = REF;
    String run = HEADS[0].RevBefore("/");
    if ( IsRegularFile(run + "/" + REF) )
        REFX = run + "/" + REF;
    else if ( IsRegularFile(workhead + "." + REF) )
        REFX = workhead + "." + REF;
    else if ( run.Contains("/") )
    {
        run = run.RevBefore("/");
        if ( IsRegularFile(run + "/" + REF) )
            REFX = run + "/" + REF;
    }
    return REFX;
}

void VAALLoadRef( const String& HEAD, const int trunc, const String& PIPELINE,
                  const String& workhead, const String& REF, String& REFX,
                  vecbasevector& ref, vec<String>& HEADS )
{    HEADS.clear();
     ExpandHead( HEAD, HEADS, trunc, PIPELINE );
     REFX = VAALRefName(REF,HEADS,workhead);
     FetchReads( ref, 0, REFX ); }

void Rebuild( vecKmerPath& unipaths, vec<tagged_rpint>& unipathsdb, 
     HyperKmerPath& h, const int K, const vecKmerPath& paths, 
     const vecKmerPath& pathsrc, const vec<tagged_rpint>& pathsdb )
{    CreateDatabase( unipaths, unipathsdb );
     digraph A;
     BuildUnipathAdjacencyGraph( paths, pathsrc, pathsdb, unipaths, unipathsdb, A );
     BuildUnipathAdjacencyHyperKmerPath( K, A, unipaths, h );
     h.RemoveUnneededVertices( ); 
     h.RemoveDeadEdgeObjects( );    
     unipaths.clear( );
     for ( int i = 0; i < h.EdgeObjectCount( ); i++ )
          unipaths.push_back_reserve( h.EdgeObject(i) );
     CreateDatabase( unipaths, unipathsdb );    }

void AssessUnipaths( const HyperKmerPath& h, const KmerBaseBroker& kbb, 
     const int KX, const basevector& ref, ostream& out )
{    out << "\nASSESSMENT OF HYPERKMERPATH\n";
     h.PrintSummaryPlus(out);
     out << "\n    " << ref.ToString( ) << endl;
     vec< triple<unsigned int,unsigned int,unsigned int> > fu;
     for ( unsigned int id = 0; id < static_cast<unsigned>(h.EdgeObjectCount( )); id++ )
     {    vec<unsigned int> f = ref.FindAll( kbb.Seq( h.EdgeObject(id) ) );
          for ( unsigned int j = 0; j < f.size( ); j++ )
               fu.push( f[j], id, f.size( ) );    }
     Sort(fu);
     for ( unsigned int j = 0; j < fu.size( ); j++ )
     {    unsigned int f = fu[j].first, id = fu[j].second;
          out << BaseAlpha(id) << " ";
          for ( int r = 0; r < 3 - BaseAlpha(id).isize( ); r++ )
               out << " ";
          for ( unsigned int r = 0; r < f; r++ )
               out << " ";
          out << kbb.Seq( h.EdgeObject(id) ).ToString( ) << " ";
          if ( fu[j].third > 1 ) out << "(M) ";
          if ( j > 0 ) out << fu[j-1].first 
               + h.EdgeObject( fu[j-1].second ).KmerCount( ) + KX - 1
               - fu[j].first;
          out << "\n";     }
     out << "\n";    }

// If an edge contains the start kmer of the reference chunk, delete
// all kmers before it.  Stop kmer: similar.

void ClipToStartStop( vecKmerPath& unipaths, const longlong x1, const longlong x2 )
{    static vec<longlong> con;
     vec<tagged_rpint> unipathsdb;
     CreateDatabase( unipaths, unipathsdb );
     Contains( unipathsdb, x1, con );
     if ( con.nonempty( ) ) 
     {    const tagged_rpint& t = unipathsdb[ con[0] ];
          KmerPath& u1 = unipaths[ t.ReadId( ) ];
          KmerPath u1new;
          u1new.AddSegment( x1, t.Stop( ) );
          for ( int j = t.PathPos( ) + 1; j < u1.NSegments( ); j++ )
               u1new.AddSegment( u1.Segment(j) );
          u1 = u1new;    }
     CreateDatabase( unipaths, unipathsdb );
     Contains( unipathsdb, x2, con );
     if ( con.nonempty( ) ) 
     {    const tagged_rpint& t = unipathsdb[ con[0] ];
          KmerPath& u2 = unipaths[ t.ReadId( ) ];
          KmerPath u2new;
          for ( int j = 0; j < t.PathPos( ); j++ )
               u2new.AddSegment( u2.Segment(j) );
          u2new.AddSegment( t.Start( ), x2 );
          u2 = u2new;    }    }

ostream& operator<<( ostream& os, const PositionLink& link ) {
  unsigned int SourceStart = link.SourcePos, SourceEnd = link.SourcePos + link.Size - 1;
  unsigned int TargetStart = link.TargetPos, TargetEnd = link.TargetPos + link.Size - 1;
  os << "[" << SourceStart << "-" << SourceEnd << "] => [" << TargetStart << "-" << TargetEnd << "]";
  return os;
}

ostream& operator<<( ostream& os, const ContigPositionLink& link ) {
  for( unsigned int tig = 0; tig < link.size(); tig++ ) {
    for( unsigned int pos = 0; pos < link[tig].size(); pos++ ) {
      const PositionLink& position = link[tig][pos];
      unsigned int SourceStart = position.SourcePos, SourceEnd = position.SourcePos + position.Size - 1;
      unsigned int TargetStart = position.TargetPos, TargetEnd = position.TargetPos + position.Size - 1;
      os << "[" << tig << "." << SourceStart << "-" << SourceEnd << "]->[" << tig << "." << TargetStart << "-" << TargetEnd << "],";
    }
  }
  return os;
}
