/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

#ifndef VAAL_POLYTOOLS_H
#define VAAL_POLYTOOLS_H

#include <vector>
#include "Basevector.h"
#include "CoreTools.h"
#include "feudal/BinaryStreamTraits.h"
#include "paths/HyperKmerPath.h"
#include "paths/KmerPath.h"
#include "paths/ReadsToPathsCoreX.h"

// Class placement: encode starting position (tig.pos) and orientation (fw) of a 
// read (id) on a genome.  Coordinates are always forward on the genome.

// WARNING: THERE IS ANOTHER CLASS OF THE SAME NAME DEFINED IN 
// PATHS/SIMULATION/PLACEMENT.H.

class placement {

     public:

     int tig;
     int pos;
     int id;
     Bool fw;

     placement( ) { }
     placement( const int tig, const int pos, const int id, const Bool fw )
          : tig(tig), pos(pos), id(id), fw(fw) { }

     friend Bool operator<( const placement& p1, const placement& p2 )
     {    if ( p1.tig < p2.tig ) return True;
          if ( p1.tig > p2.tig ) return False;
          if ( p1.pos < p2.pos ) return True;
          if ( p1.pos > p2.pos ) return False;
          if ( p1.id < p2.id ) return True;
          if ( p1.id > p2.id ) return False;
          if ( p1.fw < p2.fw ) return True;
          return False;    }

     friend Bool operator==( const placement& p1, const placement& p2 )
     {    return p1.tig == p2.tig && p1.pos == p2.pos
               && p1.id == p2.id && p1.fw == p2.fw;    }

};
TRIVIALLY_SERIALIZABLE(placement);

// AnchorReads.  Place reads on reference, anchoring using kmers that are unique
// in the reference.
//
// INPUTS:
// paths = KmerPaths generated from concatenation of reads (size = nreads)
//         and reference (size = nref)
// OUTPUT:
// places =  read placements.

void AnchorReads( const int K, const int nreads, const vecbasevector& ref,
     const vecKmerPath& paths, const vecKmerPath& pathsrc,
     const vec<big_tagged_rpint>& pathsdb, vec<placement>& places,
     const double dinuke_entropy_floor = 0.0 );

class basecall {

     public:

     basecall( ) { }
     basecall( const char base, const float qual ) : base(base), qual(qual) { }

     char base;    // 0, 1, 2, or 3
     float qual;   // four_base::CallQuality( )

     friend Bool operator<( const basecall& b1, const basecall& b2 )
     {    return b1.qual < b2.qual;    }

};

void VAALSetup( String RUN, String& HEAD, String& REF, String& PRE, 
     String& POST, String& workdir, String& workhead, int& TRUNC,
     const String& PIPELINE );

String VAALRefName( String const& REF,
                    vec<String> const& HEADS,
                    String const& workhead );

void VAALLoadRef( const String& HEAD, const int trunc, const String& PIPELINE,
                  const String& workhead, const String& REF, String& REFX,
                  vecbasevector& ref, vec<String>& HEADS );

void VAALLoad( const String& workhead, const String& HEAD, 
     const String& HEAD_TAIL, const int trunc, const String& REF, String& REFX, 
     vecbasevector& bases, vecbasevector& ref, const String& PIPELINE );

void Rebuild( vecKmerPath& unipaths, vec<tagged_rpint>& unipathsdb, 
     HyperKmerPath& h, const int K, const vecKmerPath& paths, 
     const vecKmerPath& pathsrc, const vec<tagged_rpint>& pathsdb );

void AssessUnipaths( const HyperKmerPath& h, const KmerBaseBroker& kbb, 
     const int KX, const basevector& ref, ostream& out );

// If an edge contains the start kmer of the reference chunk, delete
// all kmers before it.  Stop kmer: similar.

void ClipToStartStop( vecKmerPath& unipaths, const longlong x1, const longlong x2 );

class PositionLink {
 public:
  unsigned int SourcePos;
  unsigned int TargetPos;
  unsigned int Size;

  PositionLink() {}

  PositionLink( unsigned int _SourcePos, unsigned int _TargetPos, unsigned int _Size ) :
  SourcePos( _SourcePos ), TargetPos( _TargetPos ), Size( _Size ) {}

  unsigned int SourceStop() { return SourcePos + Size - 1; }
  unsigned int TargetStop() { return TargetPos + Size - 1; }

  Bool IsInSource( unsigned int pos ) { return pos >= SourcePos && pos < SourcePos + Size; }
  bool IsInTarget( unsigned int pos ) { return pos >= TargetPos && pos < TargetPos + Size; }

  friend Bool operator<( const PositionLink& lhs, const PositionLink& rhs ) 
  { 
    if( lhs.SourcePos != rhs.SourcePos )
      return lhs.SourcePos < rhs.SourcePos;
    return lhs.Size < rhs.Size; 
  }
  static Bool cmp_target_pos( const PositionLink& lhs, const PositionLink& rhs ) 
  { 
    if( lhs.TargetPos != rhs.TargetPos )
      return lhs.TargetPos < rhs.TargetPos;
    return lhs.Size < rhs.Size; 
  }
};

TRIVIALLY_SERIALIZABLE(PositionLink);
typedef vector< vector<PositionLink> > ContigPositionLink;

ostream& operator<<( ostream& os, const ContigPositionLink& link );
ostream& operator<<( ostream& os, const PositionLink& link );

#endif
